import unittest
import traceback

from lxml import html
from click.testing import Result, CliRunner

from glycresoft import serialize
from glycresoft.cli.analyze import (
    search_glycopeptide,
    search_glycopeptide_multipart,
)

from .fixtures import get_test_data


class _GlycopeptideSearchToolBase:
    mzml_path: str
    classic_result_db: str
    indexed_result_db: str

    classic_search_db: str = get_test_data("agp.db")

    indexed_search_db: str = get_test_data("agp_indexed.db")
    indexed_decoy_search_db: str = get_test_data("agp_indexed_decoy.db")

    indexed_gp_csv_headers = [
        "glycopeptide",
        "analysis",
        "neutral_mass",
        "mass_accuracy",
        "ms1_score",
        "ms2_score",
        "q_value",
        "total_signal",
        "start_time",
        "end_time",
        "apex_time",
        "charge_states",
        "msms_count",
        "peptide_start",
        "peptide_end",
        "protein_name",
        "mass_shifts",
        "predicted_apex_interval_start",
        "predicted_apex_interval_end",
        "retention_time_score",
        "group_id",
        "glycopeptide_score",
        "peptide_score",
        "glycan_score",
        "glycan_coverage",
        "total_q_value",
        "peptide_q_value",
        "glycan_q_value",
        "glycopeptide_q_value",
        "localizations",
        "n_glycosylation_sites",

    ]

    indexed_gpsm_csv_header = [
        "glycopeptide",
        "analysis",
        "neutral_mass",
        "mass_accuracy",
        "mass_shift_name",
        "scan_id",
        "scan_time",
        "charge",
        "ms2_score",
        "q_value",
        "precursor_abundance",
        "peptide_start",
        "peptide_end",
        "protein_name",
        "is_best_match",
        "is_precursor_fit",
        "rank",
        "group_id",
        "peptide_score",
        "glycan_score",
        "glycan_coverage",
        "peptide_q_value",
        "glycan_q_value",
        "glycopeptide_q_value",
        "localizations",
        "n_glycosylation_sites",
    ]

    gp_csv_suffix = "-glycopeptides.csv"
    gpsm_csv_suffix = "-glycopeptide-spectrum-matches.csv"
    html_report_suffix = "-report.html"

    def parse_html_report(self, html_path: str):
        tree = html.parse(html_path)
        n_prots = len(tree.findall(
            ".//section[@id='protein-table-container']/table/tbody/tr"))
        n_gps = len(tree.xpath(
            ".//*[@class and contains(concat(' ', normalize-space(@class), ' '), ' glycopeptide-detail-table-row ')]"))
        return n_prots, n_gps

    def evaluate_classic_search(self, result: Result, output_file: str):
        if result.exit_code != 0:
            print(f"Exit Code {result.exit_code}")
            print(f"Exit Code {traceback.format_exception(*result.exc_info)}")
            print(result.output)
        assert result.exit_code == 0
        db = serialize.DatabaseBoundOperation(output_file)
        ref_db = serialize.DatabaseBoundOperation(self.classic_result_db)

        for model_tp in [
            serialize.Glycopeptide,
            serialize.Peptide,
            serialize.Protein,
            serialize.GlycanComposition,
            serialize.GlycopeptideSpectrumSolutionSet,
            serialize.IdentifiedGlycopeptide,
        ]:
            assert db.query(model_tp).count() == ref_db.query(model_tp).count()

        prefix = output_file.rsplit(".", 1)[0]

        html_path = prefix + self.html_report_suffix
        n_prots, n_gps = self.parse_html_report(html_path)

        id_n_prots = db.query(serialize.Protein).join(
            serialize.Glycopeptide).join(
            serialize.IdentifiedGlycopeptide).group_by(
            serialize.Protein.id).count()
        assert n_prots == id_n_prots

        id_n_gps = db.query(serialize.IdentifiedGlycopeptide).count()
        assert id_n_gps == n_gps

    def evaluate_indexed_search(self, result: Result, output_file: str):
        if result.exit_code != 0:
            print(f"Exit Code {result.exit_code}")
            print(f"Exit Code {traceback.format_exception(*result.exc_info)}")
            print(result.output)
        assert result.exit_code == 0
        db = serialize.DatabaseBoundOperation(output_file)
        ref_db = serialize.DatabaseBoundOperation(self.indexed_result_db)

        for model_tp in [
            serialize.Glycopeptide,
            serialize.Peptide,
            serialize.Protein,
            serialize.GlycanComposition,
            serialize.GlycopeptideSpectrumMatch,
            serialize.GlycopeptideSpectrumSolutionSet,
            serialize.IdentifiedGlycopeptide,
        ]:
            assert db.query(model_tp).count() == ref_db.query(model_tp).count()

        prefix = output_file.rsplit(".", 1)[0]

        html_path = prefix + self.html_report_suffix
        n_prots, n_gps = self.parse_html_report(html_path)

        id_n_prots = db.query(serialize.Protein).join(
            serialize.Glycopeptide).join(
            serialize.IdentifiedGlycopeptide).group_by(
            serialize.Protein.id).count()
        assert n_prots == id_n_prots

        id_n_gps = db.query(serialize.IdentifiedGlycopeptide).count()
        assert id_n_gps == n_gps

        with open(prefix + self.gp_csv_suffix, 'rt') as fh:
            headers = fh.readline().strip().split(",")
            assert set(headers) <= set(self.indexed_gp_csv_headers)

        with open(prefix + self.gpsm_csv_suffix, 'rt') as fh:
            headers = fh.readline().strip().split(",")
            assert set(headers) <= set(self.indexed_gpsm_csv_header)

    def test_classic_search(self):
        runner = CliRunner()
        with runner.isolated_filesystem() as _dir_man:
            result = runner.invoke(
                search_glycopeptide,
                [
                    "-o",
                    "./classic_agp_search.db",
                    self.classic_search_db,
                    self.mzml_path,
                    "1",
                    "--export",
                    "csv",
                    "--export",
                    "html",
                    "--export",
                    "psm-csv",
                ],
            )
            self.evaluate_classic_search(result, "./classic_agp_search.db")

    def test_indexed_search(self):
        runner = CliRunner()
        with runner.isolated_filesystem() as _dir_man:
            result = runner.invoke(
                search_glycopeptide_multipart,
                [
                    "-o",
                    "./indexed_agp_search.db",
                    "-M",
                    "--export",
                    "csv",
                    "--export",
                    "html",
                    "--export",
                    "psm-csv",
                    self.indexed_search_db,
                    self.indexed_decoy_search_db,
                    self.mzml_path,
                ],
            )
            self.evaluate_indexed_search(result, "./indexed_agp_search.db")


class TestGlycopeptideSearchTool(_GlycopeptideSearchToolBase, unittest.TestCase):
    mzml_path = get_test_data("20150710_3um_AGP_001_29_30.preprocessed.mzML")
    classic_result_db = get_test_data("classic_agp_search.db")
    indexed_result_db = get_test_data("indexed_agp_search.db")


class TestGlycopeptideSearchToolNoMS2(_GlycopeptideSearchToolBase, unittest.TestCase):
    mzml_path = get_test_data("AGP_Glycomics_20150930_06.deconvoluted.mzML")
    classic_result_db = get_test_data("classic_agp_search_empty.db")
    indexed_result_db = get_test_data("indexed_agp_search_empty.db")



import os
import logging
from typing import (Any, Callable, Hashable,
                    List, Tuple, NamedTuple, Union)

from glycresoft.chromatogram_tree.mass_shift import MassShiftBase

from glycresoft.structure.structure_loader import FragmentCachingGlycopeptide

from ..spectrum_match.spectrum_match import SpectrumMatch
from ..spectrum_match.solution_set import NOParsimonyMixin


DEBUG_MODE = bool(os.environ.get('GLYCRESOFTCHROMATOGRAMDEBUG', False))

logger = logging.getLogger("glycresoft.chromatogram_mapping")
logger.addHandler(logging.NullHandler())


def always(x):
    return True


def default_threshold(x):
    return x.q_value < 0.05


Predicate = Callable[[Any], bool]
TargetType = Union[Any, FragmentCachingGlycopeptide]


class SolutionEntry(NamedTuple):
    solution: Any
    score: float
    percentile: float
    best_score: float
    match: SpectrumMatch


class NOParsimonyRepresentativeSelector(NOParsimonyMixin):
    def get_score(self, solution: SolutionEntry) -> float:
        return solution.percentile

    def get_target(self, solution: SolutionEntry) -> Any:
        return solution.match.target

    def sort(self, solution_set: List[SolutionEntry]) -> List[SolutionEntry]:
        solution_set = sorted(
            solution_set, key=lambda x: x.percentile, reverse=True)
        try:
            if solution_set and self.get_target(solution_set[0]).is_o_glycosylated():
                solution_set = self.hoist_equivalent_n_linked_solution(
                    solution_set)
        except AttributeError:
            import warnings
            warnings.warn("Could not determine glycosylation state of target of type %r" % type(
                self.get_target(solution_set[0])))
        return solution_set

    def __call__(self, solution_set: List[SolutionEntry]) -> List[SolutionEntry]:
        return self.sort(solution_set)


parsimony_sort = NOParsimonyRepresentativeSelector()


def build_glycopeptide_key(solution) -> Tuple:
    structure = solution
    key = (str(structure.clone().deglycosylate()),
           str(structure.glycan_composition))
    return key


KeyTargetMassShift = Tuple[Hashable, TargetType, MassShiftBase]
KeyMassShift = Tuple[Hashable, MassShiftBase]



from glycopeptidepy.structure.modification import (
    ModificationTarget, ModificationRule, Modification,
    ModificationCategory, SequenceLocation)


modification_rule_schema = {
    "composition": dict,
    "mono_mass": float,
    "full_name": str,
    "title": str,
    "specificity": list,
}


modification_specificity_schema = {
    "position": str,
    "site": str
}


position_map = {
    SequenceLocation.anywhere: "Anywhere",
    SequenceLocation.n_term: "Any N-term",
    SequenceLocation.c_term: "Any C-term",
}


def serialize_modification_target(target):
    payloads = []
    if not target.amino_acid_targets:
        payloads.append({
            "site": None,
            "position": position_map[target.position_modifier]
        })
        return payloads
    else:
        for aa in target.amino_acid_targets:
            payloads.append({
                "site": str(aa),
                "position": position_map[target.position_modifier]
            })
        return payloads


def serialize_modification_rule(rule):
    payload = dict()
    payload['mono_mass'] = rule.mass
    payload['full_name'] = rule.name
    payload['title'] = rule.title
    payload['composition'] = dict(rule.composition)
    payload['specificity'] = [s for t in rule.targets for s in serialize_modification_target(t)]
    return payload


def load_modification_rule(rule_dict):
    rule = ModificationRule.from_unimod(rule_dict)
    Modification.register_new_rule(rule)
    return rule


def save_modification_rule(rule):
    return serialize_modification_rule(rule)



from glypy.structure.glycan_composition import FrozenMonosaccharideResidue, HashableGlycanComposition

from glycresoft.database.glycan_composition_filter import GlycanCompositionFilter

from typing import List, Tuple

_hexose = FrozenMonosaccharideResidue.from_iupac_lite("Hex")
_hexnac = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")


def composition_distance(c1: HashableGlycanComposition, c2: HashableGlycanComposition) -> Tuple[int, float]:
    """N-Dimensional Manhattan Distance or L1 Norm"""
    keys = set(c1) | set(c2)
    distance = 0.0
    try:
        c1_get = c1._getitem_fast
        c2_get = c2._getitem_fast
    except AttributeError:
        c1_get = c1.__getitem__
        c2_get = c2.__getitem__
    for k in keys:
        distance += abs(c1_get(k) - c2_get(k))
    return int(distance), 1 / distance if distance > 0 else 1


# Old alias
n_glycan_distance = composition_distance


class CompositionSpace(object):
    filter: GlycanCompositionFilter

    def __init__(self, members):
        self.filter = GlycanCompositionFilter(members)

    @property
    def monosaccharides(self):
        return self.filter.monosaccharides

    def find_narrowly_related(self, composition: HashableGlycanComposition, window: int=1) -> List:
        partitions = []
        for i in range(len(self.monosaccharides)):
            j = 0
            m = self.monosaccharides[j]
            if i == j:
                q = self.filter.query(
                    m, composition[m] - window, composition[m] + window)
            else:
                q = self.filter.query(
                    m, composition[m], composition[m])
            for m in self.monosaccharides[1:]:
                j += 1
                center = composition[m]
                if j == i:
                    q.add(m, center - window, center + window)
                else:
                    q.add(m, center, center)
            partitions.append(q)
        out = set()
        for case in partitions:
            out.update(case)
        return out

    def l1_distance(self, c1, c2):
        keys = set(c1) | set(c2)
        distance = 0
        for k in keys:
            distance += abs(c1[k] - c2[k])
        return distance

    def find_related_broad(self, composition, window=1):
        m = self.monosaccharides[0]
        q = self.filter.query(
            m, composition[m] - window, composition[m] + window)
        for m in self.monosaccharides[1:]:
            center = composition[m]
            q.add(m, center - window, center + window)
        return q.all()

    def find_related(self, composition, window=1):
        if window == 1:
            return self.find_narrowly_related(composition, window)
        candidates = self.find_related_broad(composition, window)
        out = []
        for case in candidates:
            if self.l1_distance(composition, case) <= window:
                out.append(case)
        return out


class DistanceCache(object):
    '''A caching wrapper around a distance function (e.g. composition_distance).
    '''
    def __init__(self, distance_function):
        self.distance_function = distance_function
        self.cache = dict()

    def __call__(self, x, y):
        key = frozenset((x, y))
        try:
            return self.cache[key]
        except KeyError:
            d = self.distance_function(x, y)
            self.cache[key] = d
            return d



import csv
import warnings
from collections import defaultdict, namedtuple

from glycresoft.task import TaskBase
from glycresoft.chromatogram_tree.utils import ArithmeticMapping
from glycresoft.chromatogram_tree.chromatogram import GlycopeptideChromatogram
from glycresoft.scoring import ChromatogramSolution
from glycresoft.tandem.chromatogram_mapping import TandemAnnotatedChromatogram
from glycresoft.tandem.identified_structure import IdentifiedStructure

from glycresoft.plotting import chromatogram_artist


MergeAction = namedtuple("MergeAction", ("label", "existing", "new", "shift"))
CreateAction = namedtuple(
    "CreateAction", ("label", "structure", "chromatogram", "shift"))
LinkAction = namedtuple('LinkAction', ('label', 'from_', 'to', "shift", 'reason'))


class SharedIdentification(object):
    def __init__(self, identification_key, identifications=None, links=None):
        if identifications is None:
            identifications = dict()
        if links is None:
            links = set()
        self.identification_key = identification_key
        self.identifications = dict(identifications)
        self.links = links

    @property
    def structure(self):
        return self.identification_key

    def __getitem__(self, key):
        return self.identifications[key]

    def __setitem__(self, key, value):
        self.identifications[key] = value

    def __contains__(self, key):
        return key in self.identifications

    def __iter__(self):
        return iter(self.identifications)

    def __len__(self):
        return len(self.identifications)

    def keys(self):
        return self.identifications.keys()

    def values(self):
        return self.identifications.values()

    def items(self):
        return self.identifications.items()

    def __repr__(self):
        template = "{self.__class__.__name__}({self.identification_key}, {self.identifications})"
        return template.format(self=self)

    def weighted_neutral_masses(self):
        result = ArithmeticMapping()
        for key, value in self.identifications.items():
            result[key] = value.weighted_neutral_mass
        return result

    def apex_times(self):
        result = ArithmeticMapping()
        for key, value in self.identifications.items():
            result[key] = value.apex_time
        return result

    def total_signals(self):
        result = ArithmeticMapping()
        for key, value in self.identifications.items():
            result[key] = value.total_signal
        return result

    def plot(self, ax=None, **kwargs):
        label_map = {v.chromatogram: k for k, v in self.items()}

        def labeler(chromatogram, *args, **kwargs):
            return label_map[chromatogram]

        art = chromatogram_artist.SmoothingChromatogramArtist(
            self.values(), label_peaks=False, ax=ax, **kwargs)
        return art.draw(label_function=labeler, legend_cols=1)


class MatchBetweenRunBuilder(TaskBase):
    def __init__(self, datasets, mass_error_tolerance=1e-5, time_error_tolerance=2.0):
        self.datasets = datasets
        self.feature_table = dict()
        self.build_feature_table()
        self.labels = sorted([mbd.label for mbd in self.datasets])

        self.mass_error_tolerance = mass_error_tolerance
        self.time_error_tolerance = time_error_tolerance

    def build_feature_table(self, mass_error_tolerance=1e-5, time_error_tolerance=2.0):
        for mbd in self.datasets:
            for ids in mbd.identified_structures:
                try:
                    shared = self.feature_table[ids.structure]
                except KeyError:
                    shared = SharedIdentification(ids.structure)
                    self.feature_table[ids.structure] = shared
                if mbd.label in shared:
                    current = shared[mbd.label]
                    self.log("Multiple entries for %r in %r" %
                             (ids, mbd.label))
                    shared[mbd.label] = max(
                        [current, ids], key=lambda x: x.total_signal or 0)
                    self.log("Chose %r" % (shared[mbd.label], ))
                else:
                    shared[mbd.label] = ids

    def get_by_label(self, label):
        for mbd in self.datasets:
            if mbd.label == label:
                return mbd
        return None

    def create(self, label, structure, chromatogram, shift):
        self.log("Creating %r as %r in %r" % (chromatogram, structure, label))
        mbd = self.get_by_label(label)
        ids = mbd.create(structure, chromatogram, shift)
        if ids is not None:
            shared_id = self.feature_table[structure]
            shared_id[mbd.label] = ids
            mbd.add(ids)

    def merge(self, label, structure, new, shift):
        self.log("Merging %r into %r in %r" % (new, structure, label))
        mbd = self.get_by_label(label)
        if isinstance(new, IdentifiedStructure):
            if new == structure:
                raise ValueError("Cannot merge the same structure")
            elif new.structure.full_structure_equality(structure.structure):
                raise ValueError("Cannot merge equivalent structure")
            else:
                raise ValueError("Cannot merge ambiguous structure")
            new = new.chromatogram
        mbd.merge(structure, new, shift)

    def link(self, label, structure_from, structure_to, shift, reason):
        sf = self.feature_table[structure_from]
        st = self.feature_table[structure_to]
        link = LinkAction(label, structure_from, structure_to, shift, reason)
        sf.links.add(link)
        st.links.add(link)

    def search(self, shared_id, mass_error_tolerance=1e-5, time_error_tolerance=2.0):
        create_actions = set()
        merge_actions = set()
        link_actions = set()

        for inst in shared_id.values():
            merges, creates, links = self.find(
                inst, mass_error_tolerance, time_error_tolerance)
            create_actions.update(creates)
            merge_actions.update(merges)
            link_actions.update(links)
        return merge_actions, create_actions, link_actions

    def match_structure_between(self, shared_id, mass_error_tolerance=1e-5, time_error_tolerance=2.0):
        merge_actions, create_actions, link_actions = self.search(
            shared_id, mass_error_tolerance, time_error_tolerance)

        create_actions = sorted(
            create_actions, key=lambda x: x.chromatogram.total_signal, reverse=True)
        merge_actions = sorted(
            merge_actions, key=lambda x: x.new.total_signal, reverse=True)
        link_actions = sorted(
            link_actions, key=lambda x: max(x.from_.total_signal, x.to.total_signal), reverse=True)

        for action in create_actions:
            self.create(action.label, action.structure,
                        action.chromatogram, action.shift)

        for action in merge_actions:
            self.merge(action.label, action.existing, action.new, action.shift)

        for action in link_actions:
            self.link(action.label, action.from_.structure,
                      action.to.structure, action.shift, action.reason)

        return merge_actions, create_actions, link_actions

    def find(self, ids, mass_error_tolerance=1e-5, time_error_tolerance=2.0):
        create_actions = set()
        merge_actions = set()
        link_actions = set()

        for mbd in self.datasets:
            shared_id = self.feature_table[ids.structure]
            out = mbd.find(ids, mass_error_tolerance, time_error_tolerance)
            # We've identified this structure in this sample already
            if mbd.label in shared_id:
                existing_match = shared_id[mbd.label]
                for entity, shift in out:
                    if isinstance(entity, IdentifiedStructure):
                        if entity == existing_match:
                            continue
                        if entity.structure != existing_match.structure:
                            if entity.structure.full_structure_equality(existing_match.structure):
                                # Then is the protein different? We'll probably deal with them
                                # soon.
                                self.log("Skipping Shared Structure %r in %r\n" % (
                                    existing_match.structure, mbd.label))
                            else:
                                # Totally different structure, emit a warning?
                                warnings.warn("Ambiguous 1 Link Between %r and %r with shift %r\n" % (
                                    existing_match.structure, entity.structure, shift))
                                link_actions.add(
                                    LinkAction(mbd.label, existing_match, entity, shift, 1))
                        else:
                            if existing_match.chromatogram and entity.chromatogram:
                                self.log("Multiple chromatograms for %r at %0.2f and %0.2f\n" % (
                                    existing_match.structure, existing_match.apex_time, entity.apex_time))
                    else:
                        # It's a chromatogram, is it a different mass shift state?
                        existing_apex_time = existing_match.apex_time
                        if existing_apex_time is None:
                            existing_apex_time = existing_match.tandem_solutions[0].scan_time
                        if abs(entity.apex_time - existing_apex_time) < time_error_tolerance:
                            if existing_match.chromatogram is not None and existing_match.chromatogram.common_nodes(entity):
                                self.log("Repeated attempt to merge %r and %r in %r\n" % (
                                    existing_match.structure, entity, mbd.label))
                                continue
                            merge_actions.add(MergeAction(
                                mbd.label, existing_match.structure, entity, shift))

            else:
                for entity, shift in out:
                    if isinstance(entity, IdentifiedStructure):
                        if entity.structure.full_structure_equality(ids.structure):
                            # Then is the protein different? Should already be handled above.
                            warnings.warn("Ambiguous 2 Link Between %r and %r with shift %r in %r\n" % (
                                ids.structure, entity.structure, shift.name, mbd.label))
                        else:
                            # Totally different structure, emit a warning?
                            self.log("Ambiguous 3 Link Between %r and %r with shift %r in %r\n" % (
                                ids.structure, entity.structure, shift.name, mbd.label))
                            link_actions.add(
                                LinkAction(mbd.label, ids, entity, shift, 3))
                    else:
                        # It's a chromatogram. Wrap it in something and add it to the shared_id
                        create_actions.add(CreateAction(
                            mbd.label, ids.structure, entity, shift))

        return merge_actions, create_actions, link_actions

    def run(self):
        n = len(self.feature_table)
        for i, shared_id in enumerate(self.feature_table.values()):
            if i % 100 == 0 and i:
                self.log("Processed %d/%d shared identifications (%0.2f%%)" % (i, n, i * 100.0 / n))
            self.match_structure_between(shared_id, self.mass_error_tolerance, self.time_error_tolerance)

    def to_csv(self, stream):
        keys = ['glycopeptide', 'mass_shifts'] + [
            '%s_abundance' % a for a in self.labels] + [
            '%s_apex_time' % a for a in self.labels]
        writer = csv.DictWriter(stream, keys)
        writer.writeheader()
        for gp, shared_id in self.feature_table.items():
            mass_shifts = set()
            record = {
                "glycopeptide": str(gp)
            }
            seen = set()
            for label, inst in shared_id.items():
                record['%s_abundance' % label] = inst.total_signal or '0'
                record['%s_apex_time' % label] = inst.apex_time
                mass_shifts.update(inst.mass_shifts)
            for label in set(self.labels) - set(shared_id.keys()):
                record['%s_abundance' % label] = None
                record['%s_apex_time' % label] = None
            record['mass_shifts'] = ';'.join(m.name for m in mass_shifts)
            writer.writerow(record)



import time
import random
from glycresoft.task import TaskExecutionSequence, Pipeline

class MyTask(TaskExecutionSequence):
    def __init__(self, threshold):
        self.threshold = threshold

    def run(self, *args, **kwargs):
        while not self.error_occurred():
            self.log("I am alive!")
            time.sleep(1)
            if random.random() > self.threshold:
                self.log("Erroring out!")
                raise Exception("Failure")
        self.log("Done!")


if __name__ == "__main__":
    task = MyTask(0.2)
    task.start(True, True)
    task2 = MyTask(1.0)
    task2.start(False, True)
    pipe = Pipeline([task, task2])
    pipe.join()
    pipe.log("Stopping")
    pipe.stop()
    pipe.join(10)
    pipe.log("Stopped", task.is_alive(), pipe.error_occurred())



from glycresoft.task import TaskBase


class HypothesisSerializerBase(TaskBase):

    def set_parameters(self, params):
        if self.hypothesis.parameters is None:
            self.hypothesis.parameters = {}
            self.session.add(self.hypothesis)
            self.session.commit()
        new_params = dict(self.hypothesis.parameters)
        new_params.update(params)
        self.hypothesis.parameters = new_params
        self.session.add(self.hypothesis)
        self.session.commit()

    @property
    def hypothesis(self):
        if self._hypothesis is None:
            self._construct_hypothesis()
        return self._hypothesis

    @property
    def hypothesis_name(self):
        if self._hypothesis_name is None:
            self._construct_hypothesis()
        return self._hypothesis_name

    @property
    def hypothesis_id(self):
        if self._hypothesis_id is None:
            self._construct_hypothesis()
        return self._hypothesis_id

    def on_end(self):
        hypothesis = self.hypothesis
        self.session.add(hypothesis)
        hypothesis.status = "complete"
        self.session.add(hypothesis)
        self.session.commit()
        self.log("Hypothesis Completed")



cimport cython

from libc.stdlib cimport malloc, free, abs as labs

from collections import deque

from cpython.object cimport PyObject
from cpython.list cimport PyList_Size, PyList_GetItem, PyList_GET_ITEM
from cpython.set cimport PySet_Discard, PySet_Add
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem, PyDict_DelItem, PyDict_Items, PyDict_Values
from cpython.float cimport PyFloat_AsDouble
from cpython.int cimport PyInt_AsLong

from glypy.structure.glycan_composition import HashableGlycanComposition, FrozenMonosaccharideResidue


@cython.freelist(5000)
cdef class CompositionGraphNode(object):

    def __init__(self, composition, index, score=0., marked=False, **kwargs):
        self.composition = composition
        self.index = index
        self.edges = EdgeSet()
        self._str = str(self.composition)
        self._hash = hash(str(self._str))
        self._score = score
        self.internal_score = 0.0
        self.marked = marked

    @property
    def glycan_composition(self):
        return self.composition

    @property
    def order(self):
        return len(self.edges)

    @property
    def score(self):
        if self._score == 0:
            return self.internal_score
        else:
            return self._score

    @score.setter
    def score(self, value):
        self._score = value

    def edge_to(self, node):
        return self.edges.edge_to(self, node)

    def neighbors(self):
        result = [edge._traverse(self) for edge in self.edges]
        return result

    def __eq__(self, other):
        cdef:
            CompositionGraphNode other_typed
        if isinstance(self, CompositionGraphNode):
            if isinstance(other, CompositionGraphNode):
                other_typed = <CompositionGraphNode>other
                return self._str == other_typed._str
            else:
                return self._str == str(other)
        else:
            return str(self) == str(other)

    def __str__(self):
        return self._str

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return "CompositionGraphNode(%s, %d, %0.2f)" % (
            self._str, int(self.index) if self.index is not None else -1,
            self.score if self.score != 0 else self.internal_score)

    def copy(self):
        dup = CompositionGraphNode(self.composition, self.index, self.score)
        dup.internal_score = self.internal_score
        return dup

    def clone(self):
        return self.copy()


cdef class EdgeSet(object):

    def __init__(self, store=None):
        if store is None:
            store = dict()
        self.store = dict(store)

    def edge_to(self, CompositionGraphNode node1, CompositionGraphNode node2):
        if node2.index < node1.index:
            node1, node2 = node2, node1
        return self.store[node1, node2]

    def add(self, CompositionGraphEdge edge):
        self.store[edge.node1, edge.node2] = edge

    cpdef bint add_if_shorter(self, CompositionGraphEdge edge):
        cdef:
            object key
            PyObject* prev_ptr
            CompositionGraphEdge prev

        key = (edge.node1, edge.node2)
        prev_ptr = PyDict_GetItem(self.store, key)
        if prev_ptr == NULL:
            PyDict_SetItem(self.store, key, edge)
            return True
        else:
            prev = <CompositionGraphEdge>prev_ptr
            if prev.order < edge.order:
                return False
            else:
                PyDict_SetItem(self.store, key, edge)
                return True

    def remove(self, CompositionGraphEdge edge):
        self.store.pop((edge.node1, edge.node2))

    def __iter__(self):
        return iter(self.store.values())

    def __len__(self):
        return len(self.store)

    def __repr__(self):
        return str(set(self.store.values()))

    def __eq__(self, other):
        return self.store == other.store

    cdef list get_edges(self):
        return PyDict_Values(self.store)



@cython.freelist(15000)
cdef class CompositionGraphEdge(object):

    def __init__(self, node1, node2, order, weight=1.0):
        self.node1 = node1
        self.node2 = node2
        self.order = order if order > 0 else 1
        self.weight = weight
        self._init()

    @staticmethod
    cdef CompositionGraphEdge _create(CompositionGraphNode node1, CompositionGraphNode node2, long order, double weight):
        cdef:
            CompositionGraphEdge self

        self = CompositionGraphEdge.__new__(CompositionGraphEdge)
        self.node1 = node1
        self.node2 = node2
        self.order = order
        self.weight = weight
        self._init()
        return self

    cdef void _init(self):
        cdef:
            tuple key
        key = (self.node1, self.node2, self.order)
        self._hash = hash(key)
        self._str = "(%s, %s, %s)" % key
        self.node1.edges.add_if_shorter(self)
        self.node2.edges.add_if_shorter(self)

    def __getitem__(self, key):
        str_key = str(key)
        if str_key == self.node1._str:
            return self.node2
        elif str_key == self.node2._str:
            return self.node1
        else:
            raise KeyError(key)

    def _traverse(self, node):
        return self.node1 if node is self.node2 else self.node2

    def __eq__(self, other):
        return self._str == other._str

    def __str__(self):
        return self._str

    def __repr__(self):
        return self._str

    def __hash__(self):
        return self._hash

    def __reduce__(self):
        return self.__class__, (self.node1, self.node2, self.order, self.weight)

    cpdef copy_for(self, CompositionGraphNode node1, CompositionGraphNode node2):
        return CompositionGraphEdge._create(node1, node2, self.order, self.weight)

    def remove(self):
        try:
            self.node1.edges.remove(self)
        except KeyError:
            pass
        try:
            self.node2.edges.remove(self)
        except KeyError:
            pass


@cython.binding(True)
cpdef reindex_graph(self):
    cdef:
        size_t i, n
        CompositionGraphNode node
        list nodes

    nodes = self.nodes
    n = PyList_Size(nodes)
    for i in range(n):
        node = <CompositionGraphNode>PyList_GetItem(nodes, i)
        node.index = i


cdef double INF = float('inf')


cdef class CompositionGraphBase(object):
    cdef list get_edges(self):
        return self.edges.get_edges()

    cpdef CompositionGraphBase copy(CompositionGraphBase self):
        cdef:
            CompositionGraphNode n1, n2
            CompositionGraphEdge edge, e
            CompositionGraphBase graph
            list edges, nodes
            size_t n, i

        graph = self.__class__([], self.distance_fn)
        graph._composition_normalizer = self._composition_normalizer.copy()
        for node in self.nodes:
            graph.add_node(node.clone())

        nodes = graph.nodes
        edges = self.get_edges()

        n = PyList_Size(edges)
        for i in range(n):
            edge = <CompositionGraphEdge>PyList_GET_ITEM(edges, i)
            n1 = nodes[edge.node1.index]
            n2 = nodes[edge.node2.index]
            e = edge.copy_for(n1, n2)
            graph.edges.add(e)
        graph.neighborhoods.update(self.neighborhoods.copy())
        return graph


cdef class DijkstraPathFinder(object):

    cdef:
        public object graph
        public CompositionGraphNode start
        public CompositionGraphNode end
        public dict distance
        public dict unvisited_finite_distance
        public double limit

    def __init__(self, graph, start, end, limit=INF):
        self.graph = graph
        self.start = start
        self.end = end
        assert start._str is not None
        assert end._str is not None
        self.distance = dict()
        self.distance[start._str] = 0
        self.unvisited_finite_distance = dict()
        self.limit = limit

    cpdef set _build_initial_key_set(self):
        cdef:
            size_t i, n
            CompositionGraphNode node
            set result
            list nodes
        result = set()
        nodes = self.graph.nodes
        n = PyList_Size(nodes)
        for i in range(n):
            node = <CompositionGraphNode>PyList_GetItem(nodes, i)
            assert node._str is not None
            result.add(node._str)
        return result

    cpdef CompositionGraphNode find_smallest_unvisited(self, set unvisited):
        cdef:
            double smallest_distance, distance
            object smallest_node, key
            list iterable
            tuple entry
            PyObject* ptemp

        smallest_distance = INF
        smallest_node = None

        if not unvisited:
            raise ValueError("The unvisited set must not be empty")

        if not self.unvisited_finite_distance:
            for key in unvisited:
                ptemp = PyDict_GetItem(self.distance, key)
                if ptemp == NULL:
                    distance = INF
                if distance <= smallest_distance:
                    smallest_node = key
                    smallest_distance = distance
            if smallest_node is None:
                raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
                                 f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
        else:
            iterable = PyDict_Items(self.unvisited_finite_distance)
            for node, distance in iterable:
                if distance <= smallest_distance:
                    smallest_distance = distance
                    smallest_node = node
            if smallest_node is None:
                raise ValueError(f"Failed to select a destination from unvisited set of size {len(unvisited)} "
                                 f"with smallest distance {smallest_distance}:\n{unvisited}\nlooking for {self.start._str}->{self.end._str}")
        return self.graph[smallest_node]

    cpdef find_path(self):
        cdef:
            set unvisited
            CompositionGraphNode current_node, terminal
            EdgeSet edges
            CompositionGraphEdge edge
            PyObject* ptemp
            double path_length, terminal_distance
            long n
            int tmp

        unvisited = self._build_initial_key_set()

        visit_queue = deque([self.start])
        n = 1
        while self.end._str in unvisited:
            if n > 0:
                n -= 1
                current_node = visit_queue.popleft()
            else:
                current_node = self.find_smallest_unvisited(unvisited)

            tmp = PySet_Discard(unvisited, current_node._str)
            if tmp == -1:
                raise KeyError(current_node._str)
            elif tmp == 0:
                continue
            if PyDict_GetItem(self.unvisited_finite_distance, current_node._str) != NULL:
                PyDict_DelItem(self.unvisited_finite_distance, current_node._str)
            edges = current_node.edges
            for edge in edges:
                terminal = edge._traverse(current_node)
                if terminal._str not in unvisited:
                    continue
                ptemp = PyDict_GetItem(self.distance, current_node._str)
                if ptemp == NULL:
                    path_length = INF
                else:
                    path_length = PyFloat_AsDouble(<object>ptemp)
                path_length += edge.order
                ptemp = PyDict_GetItem(self.distance, terminal._str)
                if ptemp == NULL:
                    terminal_distance = INF
                else:
                    terminal_distance = PyFloat_AsDouble(<object>ptemp)

                if terminal_distance > path_length:
                    self.distance[terminal._str] = path_length
                    if terminal._str in unvisited and terminal_distance < self.limit:
                        self.unvisited_finite_distance[terminal._str] = path_length
                if terminal_distance < self.limit:
                    visit_queue.append(terminal)
                    n += 1

    def search(self):
        self.find_path()
        ptemp = PyDict_GetItem(self.distance, self.end._str)
        if ptemp == NULL:
            return INF
        else:
            return <object>(ptemp)


cdef int initialize_glycan_composition_vector(size_t size, glycan_composition_vector* self) nogil:
    self.size = size
    self.counts = <int*>malloc(sizeof(int) * size)
    if self.counts == NULL:
        return 1
    return 0


cdef double glycan_composition_vector_distance(glycan_composition_vector* self, glycan_composition_vector* other) nogil:
    cdef:
        size_t i, n
        double distance
    distance = 0
    n = self.size
    if other.size != n:
        return INF
    for i in range(n):
        distance += labs(self.counts[i] - other.counts[i])
    return distance


cdef int glycan_composition_vector_difference(glycan_composition_vector* self, glycan_composition_vector* other, glycan_composition_vector* into) nogil:
    cdef:
        size_t i, n

    n = self.size
    if other.size != n:
        return 1

    if into.counts == NULL:
        into.counts = <int*>malloc(sizeof(int) * n)
        into.size = n
        if into.counts == NULL:
            return 2

    for i in range(n):
        into.counts[i] = self.counts[i] - other.counts[i]
    return 0


cdef int glycan_composition_vector_addition(glycan_composition_vector* self, glycan_composition_vector* other, glycan_composition_vector* into) nogil:
    cdef:
        size_t i, n

    n = self.size
    if other.size != n:
        return 1

    if into.counts == NULL:
        into.counts = <int*>malloc(sizeof(int) * n)
        into.size = n
        if into.counts == NULL:
            return 2

    for i in range(n):
        into.counts[i] = self.counts[i] + other.counts[i]
    return 0


cdef int destroy_glycan_composition_vector(glycan_composition_vector* self) nogil:
    free(self.counts)
    return 0


cdef int copy_glycan_composition_vector(glycan_composition_vector* self, glycan_composition_vector* into) nogil:
    cdef:
        size_t i, n

    n = self.size
    if into.size != self.size:
        free(into.counts)
        into.counts = NULL

    if into.counts == NULL:
        into.counts = <int*>malloc(sizeof(int) * n)
        into.size = n
        if into.counts == NULL:
            return 1

    for i in range(n):
        into.counts[i] = self.counts[i]

    return 0


cdef class GlycanCompositionVectorContext:

    def __init__(self, components):
        self.components = list(components)
        self.component_count = len(self.components)

    def __reduce__(self):
        return self.__class__, (self.components, )

    cdef glycan_composition_vector* encode_raw(self, glycan_composition):
        cdef glycan_composition_vector* ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
        if initialize_glycan_composition_vector(self.component_count, ptr) == 1:
            raise MemoryError()
        for i in range(self.component_count):
            ptr.counts[i] = glycan_composition[self.components[i]]
        return ptr

    cpdef GlycanCompositionVector encode(self, glycan_composition):
        return GlycanCompositionVector._create(self.encode_raw(glycan_composition))

    cpdef decode(self, GlycanCompositionVector gcvec):
        cdef:
            size_t i
            int value
        gc = HashableGlycanComposition()
        if self.component_count != gcvec.ptr.size:
            raise ValueError("Dimensionality does not match")
        for i in range(self.component_count):
            value = gcvec.ptr.counts[i]
            if value != 0:
                gc[self.components[i]] = value
        return gc


cpdef _reconstruct_GlycanCompositionVector(state):
    cdef:
        glycan_composition_vector* ptr
        size_t i, n
        list counts
    counts = state[0]
    n = len(counts)
    ptr = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
    initialize_glycan_composition_vector(n, ptr)
    for i in range(n):
        ptr.counts[i] = PyInt_AsLong(counts[i])
    return GlycanCompositionVector._create(ptr)



@cython.no_gc
@cython.final
@cython.freelist(2000)
cdef class GlycanCompositionVector:

    @staticmethod
    cdef GlycanCompositionVector _create(glycan_composition_vector* ptr):
        cdef GlycanCompositionVector self = GlycanCompositionVector.__new__(GlycanCompositionVector)
        self.ptr = ptr
        return self

    def __reduce__(self):
        return _reconstruct_GlycanCompositionVector, (list(self), )

    def __dealloc__(self):
        if self.ptr != NULL:
            destroy_glycan_composition_vector(self.ptr)
            free(self.ptr)
            self.ptr = NULL

    def __getitem__(self, int i):
        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        if i >= self.ptr.size or i < 0:
            raise IndexError(i)
        return self.ptr.counts[i]

    def __setitem__(self, int i, int value):
        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        if i >= self.ptr.size or i < 0:
            raise IndexError(i)
        self.ptr.counts[i] = value

    def __len__(self):
        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        return self.ptr.size

    def __repr__(self):
        return f"{self.__class__.__name__}({[self[i] for i in range(len(self))]})"

    cpdef double distance(self, GlycanCompositionVector other) except -1:
        if other is None:
            return INF
        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        return glycan_composition_vector_distance(self.ptr, other.ptr)

    def __eq__(self, GlycanCompositionVector other):
        if other is None:
            return False
        return self.distance(other) == 0

    def __ne__(self, GlycanCompositionVector other):
        return not self == other

    cpdef GlycanCompositionVector difference(self, GlycanCompositionVector other):
        cdef:
             glycan_composition_vector* new
        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        if other is None:
            raise TypeError("Cannot compute difference from `None`")

        new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
        initialize_glycan_composition_vector(self.ptr.size, new)
        if new == NULL:
            raise MemoryError()
        if glycan_composition_vector_difference(self.ptr, other.ptr, new) == 2:
            raise MemoryError()
        return GlycanCompositionVector._create(new)

    cpdef GlycanCompositionVector addition(self, GlycanCompositionVector other):
        cdef:
             glycan_composition_vector* new
        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        if other is None:
            raise TypeError("Cannot compute addition from `None`")

        new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
        initialize_glycan_composition_vector(self.ptr.size, new)
        if new == NULL:
            raise MemoryError()
        if glycan_composition_vector_addition(self.ptr, other.ptr, new) == 2:
            raise MemoryError()
        return GlycanCompositionVector._create(new)

    def __sub__(self, GlycanCompositionVector other):
        return self.difference(other)

    def __add__(self, GlycanCompositionVector other):
        return self.addition(other)

    def __neg__(self):
        cdef:
             glycan_composition_vector* new

        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
        if new == NULL:
            raise MemoryError()
        initialize_glycan_composition_vector(self.ptr.size, new)

        for i in range(self.ptr.size):
            new.counts[i] = -self.ptr.counts[i]
        return GlycanCompositionVector._create(new)

    def __hash__(self):
        cdef:
            size_t i, n
            Py_hash_t hash_val

        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        hash_val = 0
        n = self.ptr.size
        for i in range(n):
            hash_val = (hash_val ^ self.ptr.counts[i]) * 82520UL
        return hash_val

    cpdef GlycanCompositionVector clone(self):
        cdef:
            glycan_composition_vector* new
        if self.ptr == NULL:
            raise ValueError("Object in invalid state")
        new = <glycan_composition_vector*>malloc(sizeof(glycan_composition_vector))
        if new == NULL:
            raise MemoryError()
        new.counts = NULL
        new.size = 0
        if copy_glycan_composition_vector(self.ptr, new) != 0:
            raise MemoryError()
        return GlycanCompositionVector._create(new)


cdef class CachingGlycanCompositionVectorContext(GlycanCompositionVectorContext):

    def __init__(self, components):
        super().__init__(components)
        self.encode_cache = {}
        self.decode_cache = {}

    cpdef GlycanCompositionVector encode(self, gc):
        cdef:
            PyObject* ptemp
            GlycanCompositionVector value

        ptemp = PyDict_GetItem(self.encode_cache, gc)
        if ptemp != NULL:
            return <GlycanCompositionVector>ptemp
        value = GlycanCompositionVector._create(self.encode_raw(gc))
        PyDict_SetItem(self.encode_cache, gc, value)
        PyDict_SetItem(self.decode_cache, value, gc)
        return value

    cpdef decode(self, GlycanCompositionVector gcvec):
        cdef:
            PyObject* ptemp
        ptemp = PyDict_GetItem(self.decode_cache, gcvec)
        if ptemp != NULL:
            return <object>ptemp

        value = super(CachingGlycanCompositionVectorContext, self).decode(gcvec)
        self.encode_cache[value] = gcvec
        self.decode_cache[gcvec] = value
        return value



from typing import (Callable, Collection,
                    List, Optional, Type, Union)

from ms_deisotope.peak_dependency_network.intervals import IntervalTreeNode

from glycresoft.chromatogram_tree import (
    ChromatogramFilter, Chromatogram, build_rt_interval_tree, GlycopeptideChromatogram)

from glycresoft.task import TaskBase

from ..spectrum_match import SpectrumSolutionSet

from .base import default_threshold, Predicate, DEBUG_MODE
from .revision import MS2RevisionValidator
from .chromatogram import TandemAnnotatedChromatogram, TandemSolutionsWithoutChromatogram, ScanTimeBundle
from .aggregation import GraphAnnotatedChromatogramAggregator


class ChromatogramMSMSMapper(TaskBase):
    chromatograms: List[TandemAnnotatedChromatogram]
    orphans: Union[List[ScanTimeBundle],
                   List[TandemSolutionsWithoutChromatogram]]
    error_tolerance: float
    rt_tree: IntervalTreeNode
    scan_id_to_rt: Callable[[str], float]

    def __init__(self, chromatograms: Collection[Chromatogram], error_tolerance: float=1e-5, scan_id_to_rt=lambda x: x):
        self.chromatograms = ChromatogramFilter(map(
            TandemAnnotatedChromatogram, chromatograms))
        self.rt_tree = build_rt_interval_tree(self.chromatograms)
        self.scan_id_to_rt = scan_id_to_rt
        self.orphans = []
        self.error_tolerance = error_tolerance

    def find_chromatogram_spanning(self, time: float):
        return ChromatogramFilter([interv[0] for interv in self.rt_tree.contains_point(time)])

    def find_chromatogram_for(self, solution: SpectrumSolutionSet):
        try:
            precursor_scan_time = self.scan_id_to_rt(
                solution.precursor_information.precursor_scan_id)
        except Exception:
            precursor_scan_time = self.scan_id_to_rt(solution.scan_id)
        overlapping_chroma = self.find_chromatogram_spanning(
            precursor_scan_time)
        chromas = overlapping_chroma.find_all_by_mass(
            solution.precursor_information.neutral_mass, self.error_tolerance)
        if len(chromas) == 0:
            self.orphans.append(ScanTimeBundle(solution, precursor_scan_time))
        else:
            if len(chromas) > 1:
                chroma = max(chromas, key=lambda x: x.total_signal)
            else:
                chroma = chromas[0]
            chroma.tandem_solutions.append(solution)

    def assign_solutions_to_chromatograms(self, solutions: Collection[SpectrumSolutionSet]):
        n = len(solutions)
        for i, solution in enumerate(solutions):
            if i % 5000 == 0:
                self.log("... %d/%d Solutions Handled (%0.2f%%)" %
                         (i, n, (i * 100.0 / n)))
            self.find_chromatogram_for(solution)
        if DEBUG_MODE:
            breakpoint()

    def distribute_orphans(self, threshold_fn: Predicate=default_threshold):
        lost = []
        n = len(self.orphans)
        n_chromatograms = len(self.chromatograms)
        for j, orphan in enumerate(self.orphans):
            mass = orphan.solution.precursor_ion_mass
            time = orphan.scan_time

            if j % 5000 == 0:
                self.log("... %r %d/%d Orphans Handled (%0.2f%%)" %
                         (orphan, j, n, (j * 100.0 / n)))

            candidates = self.chromatograms.find_all_by_mass(
                mass, self.error_tolerance)
            if len(candidates) > 0:
                best_index = 0
                best_distance = float('inf')
                for i, candidate in enumerate(candidates):
                    dist = min(abs(candidate.start_time - time),
                               abs(candidate.end_time - time))
                    if dist < best_distance:
                        best_index = i
                        best_distance = dist
                new_owner = candidates[best_index]

                if best_distance > 5:
                    if threshold_fn(orphan.solution):
                        lost.append(orphan.solution)
                    continue

                if DEBUG_MODE:
                    self.log("... Assigning %r to %r with %d existing solutions with distance %0.3f" %
                             (orphan, new_owner, len(new_owner.tandem_solutions), best_distance))
                new_owner.add_displaced_solution(orphan.solution)
            else:
                if threshold_fn(orphan.solution):
                    if n_chromatograms > 0 and DEBUG_MODE:
                        self.log("No chromatogram found for %r, q-value %0.4f (mass: %0.4f, time: %0.4f)" % (
                            orphan, orphan.solution.q_value, mass, time))
                    lost.append(orphan.solution)
        self.log("Distributed %d orphan identifications, %d did not find a nearby chromatogram" % (
            n, len(lost)))
        self.orphans = TandemSolutionsWithoutChromatogram.aggregate(lost)

    def assign_entities(self, threshold_fn: Predicate=default_threshold,
                        entity_chromatogram_type: Type[Chromatogram]=None):
        if entity_chromatogram_type is None:
            entity_chromatogram_type = GlycopeptideChromatogram
        for chromatogram in self:
            solutions = chromatogram.most_representative_solutions(
                threshold_fn)
            if solutions:
                solutions = sorted(
                    solutions, key=lambda x: x.score, reverse=True)
                chromatogram.assign_entity(
                    solutions[0], entity_chromatogram_type=entity_chromatogram_type)
                chromatogram.representative_solutions = solutions
        if DEBUG_MODE:
            breakpoint()

    def merge_common_entities(self, annotated_chromatograms: List[TandemAnnotatedChromatogram],
                              delta_rt: float = 0.25,
                              require_unmodified: bool = True,
                              threshold_fn: Predicate = default_threshold,
                              revision_validator: Optional[MS2RevisionValidator] = None):
        job = GraphAnnotatedChromatogramAggregator(
            annotated_chromatograms,
            delta_rt=delta_rt,
            require_unmodified=require_unmodified,
            threshold_fn=threshold_fn,
            revision_validator=revision_validator
        )
        result = job.run()
        return result

    def __len__(self):
        return len(self.chromatograms)

    def __iter__(self):
        return iter(self.chromatograms)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.chromatograms[i]
        else:
            return [self.chromatograms[j] for j in i]



from array import array
from collections import Counter

import numpy as np

from sqlalchemy import (
    Column, Numeric, Integer, String, ForeignKey,
    Table, select, join, alias, bindparam)
from sqlalchemy.orm import relationship, backref, object_session, deferred
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.ext import baked

from glycresoft.chromatogram_tree import (
    ChromatogramTreeNode as MemoryChromatogramTreeNode,
    ChromatogramTreeList,
    Chromatogram as MemoryChromatogram,
    MassShift as MemoryMassShift,
    CompoundMassShift as MemoryCompoundMassShift,
    GlycanCompositionChromatogram as MemoryGlycanCompositionChromatogram,
    ChromatogramInterface, SimpleEntityChromatogram)

from glycresoft.chromatogram_tree.chromatogram import MIN_POINTS_FOR_CHARGE_STATE
from glycresoft.chromatogram_tree.utils import ArithmeticMapping

from glycresoft.scoring import (
    ChromatogramSolution as MemoryChromatogramSolution)
from glycresoft.models import GeneralScorer

from .analysis import BoundToAnalysis
from .hypothesis import GlycanComposition

from .spectrum import (
    Base, DeconvolutedPeak, MSScan, Mass, make_memory_deconvoluted_peak)

from glypy.composition.base import formula
from glypy import Composition

try:
    trapezoid = np.trapz
except AttributeError:
    trapezoid = np.trapezoid


bakery = baked.bakery()


def extract_key(obj):
    try:
        return obj._key()
    except AttributeError:
        if obj is None:
            return obj
        return str(obj)


class SimpleSerializerCacheBase(object):
    _model_class = None

    def __init__(self, session, store=None):
        if store is None:
            store = dict()
        self.session = session
        self.store = store

    def extract_key(self, obj):
        return extract_key(obj)

    def serialize(self, obj, *args, **kwargs):
        if obj is None:
            return None
        try:
            db_obj = self.store[self.extract_key(obj)]
            return db_obj
        except KeyError:
            db_obj = self._model_class.serialize(obj, self.session, *args, **kwargs)
            self.store[self.extract_key(db_obj)] = db_obj
            return db_obj

    def __getitem__(self, obj):
        return self.serialize(obj)


class MassShift(Base):
    __tablename__ = "MassShift"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(64), index=True, unique=True)
    composition = Column(String(128))
    tandem_composition = deferred(Column(String(128)))

    _hash = None
    _mass = None
    _tandem_mass = None

    @property
    def mass(self):
        if self._mass is None:
            self._mass = Composition(str(self.composition)).mass
        return self._mass

    @property
    def tandem_mass(self):
        if self._tandem_mass is None:
            if self.tandem_composition:
                self._tandem_mass = Composition(str(self.tandem_composition)).mass
            else:
                self._tandem_mass = 0.0
        return self._tandem_mass

    def convert(self):
        try:
            tandem_composition = Composition(str(self.tandem_composition))
        except Exception:
            tandem_composition = None
        return MemoryMassShift(
            str(self.name), Composition(str(self.composition)), tandem_composition)

    @classmethod
    def serialize(cls, obj, session, *args, **kwargs):
        shift = session.query(MassShift).filter(MassShift.name == obj.name).all()
        if shift:
            return shift[0]
        else:
            db_obj = MassShift(
                name=obj.name, composition=formula(obj.composition),
                tandem_composition=formula(obj.tandem_composition))
            session.add(db_obj)
            session.flush()
            return db_obj

    def __repr__(self):
        return "DB" + repr(self.convert())

    def _key(self):
        return self.name

    def __eq__(self, other):
        return self.name == extract_key(other)

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.name)
        return self._hash


class CompoundMassShift(Base):
    __tablename__ = "CompoundMassShift"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(64), index=True, unique=True)

    counts = association_proxy(
        "_counts", "count", creator=lambda k, v: MassShiftToCompoundMassShift(individual_id=k.id, count=v))

    _mass = None
    _tandem_mass = None

    @property
    def mass(self):
        if self._mass is None:
            self._mass = sum([k.mass * v for k, v in self.counts.items()])
        return self._mass

    @property
    def tandem_mass(self):
        if self._tandem_mass is None:
            self._tandem_mass = sum([k.tandem_mass * v for k, v in self.counts.items()])
        return self._tandem_mass

    def convert(self):
        return MemoryCompoundMassShift({k.convert(): v for k, v in self.counts.items()})

    def __repr__(self):
        return "DB" + repr(self.convert())

    @classmethod
    def serialize(cls, obj, session, *args, **kwargs):
        shift = session.query(CompoundMassShift).filter(CompoundMassShift.name == obj.name).all()
        if shift:
            return shift[0]
        else:
            db_obj = CompoundMassShift(name=obj.name)
            session.add(db_obj)
            session.flush()
            if isinstance(obj, MemoryMassShift):
                db_member = MassShift.serialize(obj, session, *args, **kwargs)
                db_obj.counts[db_member] = 1
            else:
                for member, count in obj.counts.items():
                    db_member = MassShift.serialize(member, session, *args, **kwargs)
                    db_obj.counts[db_member] = count
            session.add(db_obj)
            session.flush()
            return db_obj

    def _key(self):
        return self.name


class MassShiftToCompoundMassShift(Base):
    __tablename__ = "MassShiftToCompoundMassShift"

    compound_id = Column(Integer, ForeignKey(CompoundMassShift.id, ondelete="CASCADE"), index=True, primary_key=True)
    individual_id = Column(Integer, ForeignKey(MassShift.id, ondelete='CASCADE'), index=True, primary_key=True)
    individual = relationship(MassShift)
    compound = relationship(CompoundMassShift, backref=backref(
        "_counts", collection_class=attribute_mapped_collection("individual"),
        cascade="all, delete-orphan"))

    count = Column(Integer)


class MassShiftSerializer(SimpleSerializerCacheBase):
    _model_class = CompoundMassShift

    def extract_key(self, obj):
        return obj.name


def _create_chromatogram_tree_node_branch(x):
    return ChromatogramTreeNodeBranch(child_id=x.id)


class ChromatogramTreeNode(Base, BoundToAnalysis):
    __tablename__ = "ChromatogramTreeNode"

    id = Column(Integer, primary_key=True, autoincrement=True)
    node_type_id = Column(Integer, ForeignKey(CompoundMassShift.id, ondelete='CASCADE'), index=True)
    scan_id = Column(Integer, ForeignKey(MSScan.id, ondelete='CASCADE'), index=True)
    retention_time = Column(Numeric(8, 4, asdecimal=False), index=True)
    neutral_mass = Mass()

    scan = relationship(MSScan)
    node_type = relationship(CompoundMassShift, lazy='joined')

    children = association_proxy("_children", "child", creator=_create_chromatogram_tree_node_branch)

    members = relationship(DeconvolutedPeak, secondary=lambda: ChromatogramTreeNodeToDeconvolutedPeak, lazy="subquery")

    def charge_states(self):
        u = set()
        u.update({p.charge for p in self.members})
        for child in self.children:
            u.update(child.charge_states())
        return u

    @classmethod
    def _convert(cls, session, id, node_type_cache=None, scan_id_cache=None):
        if node_type_cache is None:
            node_type_cache = dict()
        if scan_id_cache is None:
            scan_id_cache = dict()

        node_table = cls.__table__
        attribs = session.execute(node_table.select().where(node_table.c.id == id)).fetchone()

        peak_table = DeconvolutedPeak.__table__

        selector = select([DeconvolutedPeak.__table__]).select_from(
            join(peak_table, ChromatogramTreeNodeToDeconvolutedPeak)).where(
            ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == id)

        members = session.execute(selector).fetchall()

        try:
            scan_id = scan_id_cache[attribs.scan_id]
        except KeyError:
            selector = select([MSScan.__table__.c.scan_id]).where(MSScan.__table__.c.id == attribs.scan_id)
            scan_id = session.execute(selector).fetchone()
            scan_id_cache[attribs.scan_id] = scan_id

        members = [make_memory_deconvoluted_peak(m) for m in members]

        try:
            node_type = node_type_cache[attribs.node_type_id]
        except KeyError:
            shift = session.query(CompoundMassShift).get(attribs.node_type_id)
            node_type = shift.convert()
            node_type_cache[attribs.node_type_id] = node_type

        children_ids = session.query(ChromatogramTreeNodeBranch.child_id).filter(
            ChromatogramTreeNodeBranch.parent_id == id)
        children = [cls._convert(session, i[0], node_type_cache) for i in children_ids]

        return MemoryChromatogramTreeNode(
            attribs.retention_time, scan_id, children, members,
            node_type)

    def _get_child_nodes(self, session):
        query = bakery(lambda session: session.query(ChromatogramTreeNode).join(
            ChromatogramTreeNodeBranch.child).filter(
            ChromatogramTreeNodeBranch.parent_id == bindparam("parent_id")))
        result = query(session).params(parent_id=self.id).all()
        return result

    def _get_peaks(self, session):
        query = bakery(lambda session: session.query(DeconvolutedPeak).join(
            ChromatogramTreeNodeToDeconvolutedPeak).filter(
            ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == bindparam("node_id")))
        result = query(session).params(node_id=self.id).all()
        return result

    def convert(self, node_type_cache=None, scan_id_cache=None):
        if scan_id_cache is not None:
            try:
                scan_id = scan_id_cache[self.scan_id]
            except KeyError:
                scan_id = self.scan.scan_id
                scan_id_cache[self.scan_id] = scan_id
        else:
            scan_id = self.scan.scan_id
        if node_type_cache is not None:
            try:
                node_type = node_type_cache[self.node_type_id]
            except KeyError:
                node_type = self.node_type.convert()
                node_type_cache[self.node_type_id] = node_type
        else:
            node_type = self.node_type.convert()

        session = object_session(self)
        # children = self.children
        children = self._get_child_nodes(session)
        # peaks = self.members
        peaks = self._get_peaks(session)

        inst = MemoryChromatogramTreeNode(
            self.retention_time, scan_id, [
                child.convert(node_type_cache, scan_id_cache) for child in children],
            [p.convert() for p in peaks],
            node_type)
        return inst

    @classmethod
    def serialize(cls, obj, session, analysis_id, peak_lookup_table=None, mass_shift_cache=None,
                  scan_lookup_table=None, node_peak_map=None, *args, **kwargs):
        if mass_shift_cache is None:
            mass_shift_cache = MassShiftSerializer(session)
        inst = ChromatogramTreeNode(
            scan_id=scan_lookup_table[obj.scan_id], retention_time=obj.retention_time,
            analysis_id=analysis_id)
        nt = mass_shift_cache.serialize(obj.node_type)
        inst.node_type_id = nt.id

        session.add(inst)
        session.flush()

        if peak_lookup_table is not None:
            member_ids = []
            blocked = 0
            for member in obj.members:
                peak_id = peak_lookup_table[obj.scan_id, member]
                node_peak_key = (inst.id, peak_id)
                if node_peak_key in node_peak_map:
                    blocked += 1
                    continue
                node_peak_map[node_peak_key] = True
                member_ids.append(peak_id)

            if len(member_ids):
                session.execute(ChromatogramTreeNodeToDeconvolutedPeak.insert(), [  # pylint: disable=no-value-for-parameter
                    {'node_id': inst.id, 'peak_id': member_id} for member_id in member_ids])
            elif blocked == 0:
                raise Exception("No Peaks Saved")

        children = [cls.serialize(
            child, session,
            analysis_id=analysis_id,
            peak_lookup_table=peak_lookup_table,
            mass_shift_cache=mass_shift_cache,
            scan_lookup_table=scan_lookup_table,
            node_peak_map=node_peak_map,
            *args, **kwargs) for child in obj.children]
        branches = [
            ChromatogramTreeNodeBranch(parent_id=inst.id, child_id=child.id)
            for child in children
        ]
        assert len(branches) == len(obj.children)
        session.add_all(branches)
        session.add(inst)
        # assert len(inst.children) == len(obj.children)
        session.flush()
        return inst

    def _total_intensity_query(self, session):
        session.query()


class ChromatogramTreeNodeBranch(Base):
    __tablename__ = "ChromatogramTreeNodeBranch"

    parent_id = Column(Integer, ForeignKey(ChromatogramTreeNode.id, ondelete="CASCADE"), index=True, primary_key=True)
    child_id = Column(Integer, ForeignKey(ChromatogramTreeNode.id, ondelete="CASCADE"), index=True, primary_key=True)
    child = relationship(ChromatogramTreeNode, backref=backref(
        "parent"), foreign_keys=[child_id])
    parent = relationship(ChromatogramTreeNode, backref=backref(
        "_children", lazy='subquery'),
        foreign_keys=[parent_id])

    def __repr__(self):
        return "ChromatogramTreeNodeBranch(%r, %r)" % (self.parent_id, self.child_id)


ChromatogramTreeNodeToDeconvolutedPeak = Table(
    "ChromatogramTreeNodeToDeconvolutedPeak", Base.metadata,
    Column("node_id", Integer, ForeignKey(
        ChromatogramTreeNode.id, ondelete='CASCADE'), primary_key=True),
    Column("peak_id", Integer, ForeignKey(
        DeconvolutedPeak.id, ondelete="CASCADE"), primary_key=True))


class Chromatogram(Base, BoundToAnalysis):
    __tablename__ = "Chromatogram"

    _total_signal_cache = None

    id = Column(Integer, primary_key=True, autoincrement=True)
    neutral_mass = Mass()

    @property
    def composition(self):
        return None

    start_time = Column(Numeric(8, 4, asdecimal=False), index=True)
    end_time = Column(Numeric(8, 4, asdecimal=False), index=True)

    nodes = relationship(ChromatogramTreeNode, secondary=lambda: ChromatogramToChromatogramTreeNode)

    @property
    def mass_shifts(self):
        session = object_session(self)
        return self._mass_shifts_query(session)

    def _mass_shifts_query(self, session):
        # (root_peaks_join, branch_peaks_join,
        #  anode, bnode, apeak) = _mass_shifts_query_inner

        # branch_node_info = select([anode.c.node_type_id, anode.c.retention_time]).where(
        #     ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id
        # ).select_from(branch_peaks_join)

        # root_node_info = select([anode.c.node_type_id, anode.c.retention_time]).where(
        #     ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id
        # ).select_from(root_peaks_join)

        # all_node_info_q = root_node_info.union_all(branch_node_info).order_by(
        #     anode.c.retention_time)

        # all_node_info = session.execute(all_node_info_q).fetchall()

        all_node_info = session.execute(_mass_shift_query_stmt, dict(id=self.id)).fetchall()

        node_type_ids = set()
        for node_type_id, rt in all_node_info:
            node_type_ids.add(node_type_id)

        node_types = []
        for ntid in node_type_ids:
            node_types.append(session.query(CompoundMassShift).get(ntid).convert())
        return node_types

    def mass_shift_signal_fractions(self):
        session = object_session(self)
        return self._mass_shift_signal_fraction_query(session)

    def bisect_mass_shift(self, mass_shift):
        session = object_session(self)

        (root_peaks_join, branch_peaks_join,
         anode, bnode, apeak) = _mass_shifts_query_inner
        root_node_info = select(
            [anode.c.node_type_id, anode.c.retention_time, apeak.c.intensity]).where(
            ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id
        ).select_from(root_peaks_join)

        branch_node_info = select(
            [anode.c.node_type_id, anode.c.retention_time, apeak.c.intensity]).where(
            ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id
        ).select_from(branch_peaks_join)

        all_node_info_q = root_node_info.union_all(branch_node_info).order_by(anode.c.retention_time)
        mass_shift_id = session.query(CompoundMassShift.id).filter(CompoundMassShift.name == mass_shift.name).scalar()

        new_mass_shift = SimpleEntityChromatogram()
        new_no_mass_shift = SimpleEntityChromatogram()

        for r in sorted(session.execute(all_node_info_q), key=lambda x: x.retention_time):
            if r.node_type_id == mass_shift_id:
                new_mass_shift.setdefault(r.retention_time, 0)
                new_mass_shift[r.retention_time] += r.intensity
            else:
                new_no_mass_shift.setdefault(r.retention_time, 0)
                new_no_mass_shift[r.retention_time] += r.intensity
        new_mass_shift.mass_shifts = {mass_shift}
        new_no_mass_shift.mass_shifts = set(self.mass_shifts) - {mass_shift}

        return new_mass_shift, new_no_mass_shift

    def _mass_shift_signal_fraction_query(self, session):
        (root_peaks_join, branch_peaks_join,
         anode, bnode, apeak) = _mass_shifts_query_inner
        root_node_info = select([anode.c.node_type_id, anode.c.retention_time, apeak.c.intensity]).where(
            ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id
        ).select_from(root_peaks_join)

        branch_node_info = select([anode.c.node_type_id, anode.c.retention_time, apeak.c.intensity]).where(
            ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id
        ).select_from(branch_peaks_join)

        all_node_info_q = root_node_info.union_all(branch_node_info).order_by(anode.c.retention_time)
        acc = ArithmeticMapping()
        for r in session.execute(all_node_info_q):
            acc[r.node_type_id] += r.intensity

        acc = ArithmeticMapping(
            {(session.query(CompoundMassShift).get(ntid).convert()): signal
             for ntid, signal in acc.items()})
        return acc

    def get_chromatogram(self):
        return self

    def _peaks_query(self, session):
        anode = alias(ChromatogramTreeNode.__table__)
        bnode = alias(ChromatogramTreeNode.__table__)
        apeak = alias(DeconvolutedPeak.__table__)

        peak_join = apeak.join(
            ChromatogramTreeNodeToDeconvolutedPeak,
            ChromatogramTreeNodeToDeconvolutedPeak.c.peak_id == apeak.c.id)

        root_peaks_join = peak_join.join(
            anode,
            ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == anode.c.id).join(
            ChromatogramToChromatogramTreeNode,
            ChromatogramToChromatogramTreeNode.c.node_id == anode.c.id)

        branch_peaks_join = peak_join.join(
            anode,
            ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == anode.c.id).join(
            ChromatogramTreeNodeBranch,
            ChromatogramTreeNodeBranch.child_id == anode.c.id).join(
            bnode, ChromatogramTreeNodeBranch.parent_id == bnode.c.id).join(
            ChromatogramToChromatogramTreeNode,
            ChromatogramToChromatogramTreeNode.c.node_id == bnode.c.id)

        branch_ids = select([apeak.c.id]).where(
            ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id
        ).select_from(branch_peaks_join)

        root_ids = select([apeak.c.id]).where(
            ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id
        ).select_from(root_peaks_join)

        all_ids = root_ids.union_all(branch_ids)
        peaks = session.execute(all_ids).fetchall()
        return {p[0] for p in peaks}

    _peak_hash_ = None

    def _build_peak_hash(self):
        if self._peak_hash_ is None:
            session = object_session(self)
            self._peak_hash_ = frozenset(self._peaks_query(session))
        return self._peak_hash_

    @property
    def _peak_hash(self):
        return self._build_peak_hash()

    def is_distinct(self, other):
        return self._peak_hash.isdisjoint(other._peak_hash)

    @property
    def charge_states(self):
        states = Counter()
        for node in self.nodes:
            states += (Counter(node.charge_states()))
        # Require more than `MIN_POINTS_FOR_CHARGE_STATE` data points to accept any
        # charge state
        collapsed_states = {k for k, v in states.items() if v >= min(MIN_POINTS_FOR_CHARGE_STATE, len(self))}
        if not collapsed_states:
            collapsed_states = set(states.keys())
        return collapsed_states

    def raw_convert(self, node_type_cache=None, scan_id_cache=None):
        session = object_session(self)
        node_ids = session.query(ChromatogramToChromatogramTreeNode.c.node_id).filter(
            ChromatogramToChromatogramTreeNode.c.chromatogram_id == self.id).all()
        nodes = [ChromatogramTreeNode._convert(
            session, ni[0], node_type_cache=node_type_cache,
            scan_id_cache=scan_id_cache) for ni in node_ids]
        nodes.sort(key=lambda x: x.retention_time)
        inst = MemoryChromatogram(None, ChromatogramTreeList(nodes))
        return inst

    def orm_convert(self, *args, **kwargs):
        nodes = [node.convert(*args, **kwargs) for node in self.nodes]
        nodes.sort(key=lambda x: x.retention_time)
        inst = MemoryChromatogram(None, ChromatogramTreeList(nodes))
        return inst

    def convert(self, node_type_cache=None, scan_id_cache=None, **kwargs):
        return self.orm_convert(
            node_type_cache=node_type_cache,
            scan_id_cache=scan_id_cache, **kwargs)

    @classmethod
    def serialize(cls, obj, session, analysis_id, peak_lookup_table=None, mass_shift_cache=None,
                  scan_lookup_table=None, node_peak_map=None, *args, **kwargs):
        if mass_shift_cache is None:
            mass_shift_cache = MassShiftSerializer(session)
        inst = cls(
            neutral_mass=obj.neutral_mass, start_time=obj.start_time, end_time=obj.end_time,
            analysis_id=analysis_id)
        session.add(inst)
        session.flush()
        node_ids = []
        for node in obj.nodes:
            db_node = ChromatogramTreeNode.serialize(
                node, session, analysis_id=analysis_id,
                peak_lookup_table=peak_lookup_table,
                mass_shift_cache=mass_shift_cache,
                scan_lookup_table=scan_lookup_table,
                node_peak_map=node_peak_map, *args, **kwargs)
            node_ids.append(db_node.id)
        session.execute(ChromatogramToChromatogramTreeNode.insert(), [  # pylint: disable=no-value-for-parameter
            {"chromatogram_id": inst.id, "node_id": node_id} for node_id in node_ids])
        return inst

    def _weighted_neutral_mass_query(self, session):
        all_intensity_mass = session.execute(_weighted_neutral_mass_stmt, dict(id=self.id)).fetchall()
        shift_ids = {}
        acc_mass = 0.0
        acc_intensity = 0.0
        shift_ids = {}
        for intensity, mass, shift_id in all_intensity_mass:
            try:
                delta_mass = shift_ids[shift_id]
            except KeyError:
                delta_mass = shift_ids[shift_id] = session.query(CompoundMassShift).get(shift_id).convert().mass
            acc_mass += (mass - delta_mass) * intensity
            acc_intensity += intensity
        return acc_mass / acc_intensity

    def _as_array_query(self, session):
        all_intensities = session.execute(_as_arrays_stmt, dict(id=self.id)).fetchall()

        time = array('d')
        signal = array('d')
        current_signal = all_intensities[0][0]
        current_time = all_intensities[0][1]

        for intensity, rt in all_intensities[1:]:
            if abs(current_time - rt) < 1e-4:
                current_signal += intensity
            else:
                time.append(current_time)
                signal.append(current_signal)
                current_time = rt
                current_signal = intensity
        time.append(current_time)
        signal.append(current_signal)
        # Make sure intensity dtype doesn't undergo signed overflow by being
        # sufficiently wide to hold the value (e.g. not int32)
        return np.array(time), np.array(signal, dtype=np.float64)

    def as_arrays(self):
        session = object_session(self)
        return self._as_array_query(session)

    @property
    def weighted_neutral_mass(self):
        session = object_session(self)
        return self._weighted_neutral_mass_query(session)

    def integrate(self):
        time, intensity = self.as_arrays()
        return trapezoid(intensity, time)

    @property
    def total_signal(self):
        if self._total_signal_cache is None:
            session = object_session(self)
            self._total_signal_cache = self._as_array_query(session)[1].sum()
        return self._total_signal_cache

    @property
    def apex_time(self):
        time, intensity = self.as_arrays()
        return time[np.argmax(intensity)]

    def __repr__(self):
        return "DB" + repr(self.convert())

    def __len__(self):
        return len(self.nodes)

    def __getitem__(self, i):
        return self.nodes[i]

    def __iter__(self):
        return iter(self.nodes)


ChromatogramInterface.register(Chromatogram)


class MissingChromatogramError(ValueError):
    pass


class ChromatogramWrapper(object):

    def _get_chromatogram(self):
        if self.chromatogram is None:
            raise MissingChromatogramError()
        return self.chromatogram

    def has_chromatogram(self):
        return self.chromatogram is None

    def integrate(self):
        return self._get_chromatogram().integrate()

    @property
    def _peak_hash(self):
        try:
            return self._get_chromatogram()._peak_hash
        except MissingChromatogramError:
            return frozenset()

    def is_distinct(self, other):
        return self._peak_hash.isdisjoint(other._peak_hash)

    @property
    def mass_shifts(self):
        try:
            return self._get_chromatogram().mass_shifts
        except MissingChromatogramError:
            return []

    def mass_shift_signal_fractions(self):
        try:
            return self._get_chromatogram().mass_shift_signal_fractions()
        except MissingChromatogramError:
            return ArithmeticMapping()

    @property
    def weighted_neutral_mass(self):
        return self._get_chromatogram().weighted_neutral_mass

    @property
    def total_signal(self):
        return self._get_chromatogram().total_signal

    def __len__(self):
        return len(self._get_chromatogram())

    def as_arrays(self):
        return self._get_chromatogram().as_arrays()

    @property
    def start_time(self):
        return self._get_chromatogram().start_time

    @property
    def end_time(self):
        return self._get_chromatogram().end_time

    @property
    def apex_time(self):
        return self._get_chromatogram().apex_time

    @property
    def neutral_mass(self):
        return self._get_chromatogram().neutral_mass

    @property
    def charge_states(self):
        return self._get_chromatogram().charge_states

    def bisect_mass_shift(self, mass_shift):
        return self._get_chromatogram().bisect_mass_shift(mass_shift)


ChromatogramInterface.register(ChromatogramWrapper)


ChromatogramToChromatogramTreeNode = Table(
    "ChromatogramToChromatogramTreeNode", Base.metadata,
    Column("chromatogram_id", Integer, ForeignKey(
        Chromatogram.id, ondelete="CASCADE"), primary_key=True),
    Column("node_id", Integer, ForeignKey(
        ChromatogramTreeNode.id, ondelete='CASCADE'), primary_key=True))


class CompositionGroup(Base, BoundToAnalysis):
    __tablename__ = "CompositionGroup"

    id = Column(Integer, primary_key=True)
    composition = Column(String(512), index=True)

    @classmethod
    def serialize(cls, obj, session, *args, **kwargs):
        if obj is None:
            return None
        composition = session.query(cls).filter(cls.composition == str(obj)).all()
        if composition:
            return composition[0]
        else:
            db_obj = cls(composition=str(obj))
            session.add(db_obj)
            session.flush()
            return db_obj

    def convert(self):
        return str(self.composition)

    def _key(self):
        return self.composition

    def __repr__(self):
        return "CompositionGroup(%r, %d)" % (self.convert(), self.id)


class CompositionGroupSerializer(SimpleSerializerCacheBase):
    _model_class = CompositionGroup


class ScoredChromatogram(object):
    score = Column(Numeric(8, 7, asdecimal=False), index=True)
    internal_score = Column(Numeric(8, 7, asdecimal=False))


class ChromatogramSolution(Base, BoundToAnalysis, ScoredChromatogram, ChromatogramWrapper):
    __tablename__ = "ChromatogramSolution"

    id = Column(Integer, primary_key=True)

    chromatogram_id = Column(Integer, ForeignKey(
        Chromatogram.id, ondelete='CASCADE'), index=True)
    composition_group_id = Column(Integer, ForeignKey(
        CompositionGroup.id, ondelete='CASCADE'), index=True)

    chromatogram = relationship(Chromatogram, lazy='select')
    composition_group = relationship(CompositionGroup)

    def get_chromatogram(self):
        return self.chromatogram.get_chromatogram()

    @property
    def key(self):
        if self.composition_group is not None:
            return self.composition_group._key()
        else:
            return self.neutral_mass

    @property
    def composition(self):
        return self.composition_group

    def convert(self, *args, **kwargs):
        chromatogram_scoring_model = kwargs.pop(
            "chromatogram_scoring_model", GeneralScorer)
        chromatogram = self.chromatogram.convert(*args, **kwargs)
        composition = self.composition_group.convert() if self.composition_group else None
        chromatogram.composition = composition
        sol = MemoryChromatogramSolution(
            chromatogram, self.score, chromatogram_scoring_model, self.internal_score)
        used_as_mass_shift = []
        for pair in self.used_as_mass_shift:
            used_as_mass_shift.append((pair[0], pair[1].convert()))
        sol.used_as_mass_shift = used_as_mass_shift

        ambiguous_with = []
        for pair in self.ambiguous_with:
            ambiguous_with.append((pair[0], pair[1].convert()))
        sol.ambiguous_with = ambiguous_with
        return sol

    @property
    def used_as_mass_shift(self):
        pairs = []
        for rel in self._mass_shifted_relationships_mass_shift:
            pairs.append((rel.owner.key, rel.mass_shift))
        return pairs

    @property
    def ambiguous_with(self):
        pairs = []
        for rel in self._mass_shift_relationships_owned:
            pairs.append((rel.mass_shifted.key, rel.mass_shift))
        return pairs

    @classmethod
    def serialize(cls, obj, session, analysis_id, peak_lookup_table=None, mass_shift_cache=None,
                  scan_lookup_table=None, composition_cache=None, node_peak_map=None,
                  *args, **kwargs):
        if mass_shift_cache is None:
            mass_shift_cache = MassShiftSerializer(session)
        if composition_cache is None:
            composition_cache = CompositionGroupSerializer(session)
        db_composition_group = composition_cache.serialize(obj.composition)
        composition_group_id = db_composition_group.id if db_composition_group is not None else None
        inst = cls(
            composition_group_id=composition_group_id,
            analysis_id=analysis_id, score=obj.score,
            internal_score=obj.internal_score)
        chromatogram = Chromatogram.serialize(
            obj.chromatogram, session=session, analysis_id=analysis_id,
            peak_lookup_table=peak_lookup_table, mass_shift_cache=mass_shift_cache,
            scan_lookup_table=scan_lookup_table, node_peak_map=node_peak_map,
            *args, **kwargs)
        inst.chromatogram_id = chromatogram.id
        session.add(inst)
        session.flush()
        return inst

    def __repr__(self):
        return "DB" + repr(self.convert())


class ChromatogramSolutionWrapper(ChromatogramWrapper):
    def _get_chromatogram(self):
        return self.solution

    @property
    def used_as_mass_shift(self):
        return self._get_chromatogram().used_as_mass_shift

    @property
    def ambiguous_with(self):
        return self._get_chromatogram().ambiguous_with

    def bisect_mass_shift(self, mass_shift):
        chromatogram = self._get_chromatogram()
        new_mass_shift, new_no_mass_shift = chromatogram.bisect_mass_shift(mass_shift)

        for chrom in (new_mass_shift, new_no_mass_shift):
            chrom.entity = self.entity
            chrom.composition = self.entity
            chrom.glycan_composition = self.glycan_composition
        return new_mass_shift, new_no_mass_shift


class GlycanCompositionChromatogram(Base, BoundToAnalysis, ScoredChromatogram, ChromatogramSolutionWrapper):
    __tablename__ = "GlycanCompositionChromatogram"

    id = Column(Integer, primary_key=True)

    chromatogram_solution_id = Column(
        Integer, ForeignKey(ChromatogramSolution.id, ondelete='CASCADE'), index=True)

    solution = relationship(ChromatogramSolution)

    glycan_composition_id = Column(
        Integer, ForeignKey(GlycanComposition.id, ondelete='CASCADE'), index=True)

    entity = relationship(GlycanComposition)

    def convert(self, *args, **kwargs):
        entity = self.entity.convert()
        solution = self.solution.convert(*args, **kwargs)
        case = solution.chromatogram.clone(MemoryGlycanCompositionChromatogram)
        case.composition = entity
        solution.chromatogram = case

        solution.id = self.id
        solution.solution_id = self.solution.id

        try:
            solution.tandem_solutions = self.spectrum_cluster.convert()
        except AttributeError:
            solution.tandem_solutions = []

        return solution

    @classmethod
    def serialize(cls, obj, session, analysis_id, peak_lookup_table=None, mass_shift_cache=None,
                  scan_lookup_table=None, composition_cache=None, node_peak_map=None,
                  *args, **kwargs):
        solution = ChromatogramSolution.serialize(
            obj, session, analysis_id, peak_lookup_table=peak_lookup_table,
            mass_shift_cache=mass_shift_cache, scan_lookup_table=scan_lookup_table,
            composition_cache=composition_cache, node_peak_map=node_peak_map,
            *args, **kwargs)
        inst = cls(
            chromatogram_solution_id=solution.id, glycan_composition_id=obj.composition.id,
            score=obj.score, internal_score=obj.internal_score, analysis_id=analysis_id)
        session.add(inst)
        session.flush()
        return inst

    @property
    def glycan_composition(self):
        return self.entity.convert()

    @property
    def composition(self):
        return self.glycan_composition

    @property
    def key(self):
        return self.glycan_composition

    def __repr__(self):
        return "DB" + repr(self.convert())


class UnidentifiedChromatogram(Base, BoundToAnalysis, ScoredChromatogram, ChromatogramSolutionWrapper):
    __tablename__ = "UnidentifiedChromatogram"

    id = Column(Integer, primary_key=True)

    chromatogram_solution_id = Column(
        Integer, ForeignKey(ChromatogramSolution.id, ondelete='CASCADE'), index=True)

    solution = relationship(ChromatogramSolution)

    @property
    def key(self):
        return self.neutral_mass

    def convert(self, *args, **kwargs):
        solution = self.solution.convert(*args, **kwargs)
        solution.id = self.id
        solution.solution_id = self.solution.id

        try:
            solution.tandem_solutions = self.spectrum_cluster.convert()
        except AttributeError:
            solution.tandem_solutions = []

        return solution

    @classmethod
    def serialize(cls, obj, session, analysis_id, peak_lookup_table=None, mass_shift_cache=None,
                  scan_lookup_table=None, composition_cache=None, node_peak_map=None,
                  *args, **kwargs):
        solution = ChromatogramSolution.serialize(
            obj, session, analysis_id, peak_lookup_table,
            mass_shift_cache, scan_lookup_table, composition_cache,
            node_peak_map=node_peak_map, *args, **kwargs)
        inst = cls(
            chromatogram_solution_id=solution.id,
            score=obj.score, internal_score=obj.internal_score,
            analysis_id=analysis_id)
        session.add(inst)
        session.flush()
        return inst

    def __repr__(self):
        return "DBUnidentified" + repr(self.convert())

    @property
    def glycan_composition(self):
        return None

    @property
    def composition(self):
        return None


class ChromatogramSolutionMassShiftedToChromatogramSolution(Base):
    __tablename__ = "ChromatogramSolutionAdductToChromatogramSolution"

    mass_shifted_solution_id = Column(Integer, ForeignKey(
        ChromatogramSolution.id, ondelete="CASCADE"), primary_key=True)
    mass_shift_id = Column(Integer, ForeignKey(
        CompoundMassShift.id, ondelete='CASCADE'), primary_key=True)
    owning_solution_id = Column(Integer, ForeignKey(
        ChromatogramSolution.id, ondelete="CASCADE"), primary_key=True)

    owner = relationship(ChromatogramSolution, backref=backref(
        "_mass_shift_relationships_owned"),
        foreign_keys=[owning_solution_id])
    mass_shifted = relationship(ChromatogramSolution, backref=backref(
        "_mass_shifted_relationships_mass_shift"),
        foreign_keys=[mass_shifted_solution_id])
    mass_shift = relationship(CompoundMassShift)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.mass_shifted.key} -{self.mass_shift.name}-> {self.owner.key})"
        return template.format(self=self)


# SQL Operation pre-definition. The Chromatogram table involves several complex joins to load
# relevant tree-like structure information. To reduce the amount of time SQLAlchemy spends
# on compiling these queries into strings, we use ``bindparam``-instrumented statements
# and expressions. These statements are made global.


def __build_mass_shifts_query_inner():

    anode = alias(ChromatogramTreeNode.__table__)
    bnode = alias(ChromatogramTreeNode.__table__)
    apeak = alias(DeconvolutedPeak.__table__)

    peak_join = apeak.join(
        ChromatogramTreeNodeToDeconvolutedPeak,
        ChromatogramTreeNodeToDeconvolutedPeak.c.peak_id == apeak.c.id)

    root_peaks_join = peak_join.join(
        anode,
        ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == anode.c.id).join(
        ChromatogramToChromatogramTreeNode,
        ChromatogramToChromatogramTreeNode.c.node_id == anode.c.id)

    branch_peaks_join = peak_join.join(
        anode,
        ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == anode.c.id).join(
        ChromatogramTreeNodeBranch,
        ChromatogramTreeNodeBranch.child_id == anode.c.id).join(
        bnode, ChromatogramTreeNodeBranch.parent_id == bnode.c.id).join(
        ChromatogramToChromatogramTreeNode,
        ChromatogramToChromatogramTreeNode.c.node_id == bnode.c.id)
    return root_peaks_join, branch_peaks_join, anode, bnode, apeak


_mass_shifts_query_inner = __build_mass_shifts_query_inner()


def __build_mass_shifts_query():
    (root_peaks_join, branch_peaks_join,
        anode, bnode, apeak) = _mass_shifts_query_inner

    branch_node_info = select([anode.c.node_type_id, anode.c.retention_time]).where(
        ChromatogramToChromatogramTreeNode.c.chromatogram_id == bindparam('id')
    ).select_from(branch_peaks_join)

    root_node_info = select([anode.c.node_type_id, anode.c.retention_time]).where(
        ChromatogramToChromatogramTreeNode.c.chromatogram_id == bindparam('id')
    ).select_from(root_peaks_join)

    all_node_info_q = root_node_info.union_all(branch_node_info).order_by(
        anode.c.retention_time)
    return all_node_info_q


_mass_shift_query_stmt = __build_mass_shifts_query()


def __build_weighted_neutral_mass_query():
    anode = alias(ChromatogramTreeNode.__table__)
    bnode = alias(ChromatogramTreeNode.__table__)
    apeak = alias(DeconvolutedPeak.__table__)

    peak_join = apeak.join(
        ChromatogramTreeNodeToDeconvolutedPeak,
        ChromatogramTreeNodeToDeconvolutedPeak.c.peak_id == apeak.c.id)

    root_peaks_join = peak_join.join(
        anode,
        ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == anode.c.id).join(
        ChromatogramToChromatogramTreeNode,
        ChromatogramToChromatogramTreeNode.c.node_id == anode.c.id)

    branch_peaks_join = peak_join.join(
        anode,
        ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == anode.c.id).join(
        ChromatogramTreeNodeBranch,
        ChromatogramTreeNodeBranch.child_id == anode.c.id).join(
        bnode, ChromatogramTreeNodeBranch.parent_id == bnode.c.id).join(
        ChromatogramToChromatogramTreeNode,
        ChromatogramToChromatogramTreeNode.c.node_id == bnode.c.id)

    branch_intensities = select([apeak.c.intensity, apeak.c.neutral_mass, anode.c.node_type_id]).where(
        ChromatogramToChromatogramTreeNode.c.chromatogram_id == bindparam("id")
    ).select_from(branch_peaks_join)

    root_intensities = select([apeak.c.intensity, apeak.c.neutral_mass, anode.c.node_type_id]).where(
        ChromatogramToChromatogramTreeNode.c.chromatogram_id == bindparam("id")
    ).select_from(root_peaks_join)

    all_intensity_mass_q = root_intensities.union_all(branch_intensities)
    return all_intensity_mass_q


_weighted_neutral_mass_stmt = __build_weighted_neutral_mass_query()


def __build_as_arrays_query():
    anode = alias(ChromatogramTreeNode.__table__)
    bnode = alias(ChromatogramTreeNode.__table__)
    apeak = alias(DeconvolutedPeak.__table__)

    peak_join = apeak.join(
        ChromatogramTreeNodeToDeconvolutedPeak,
        ChromatogramTreeNodeToDeconvolutedPeak.c.peak_id == apeak.c.id)

    root_peaks_join = peak_join.join(
        anode,
        ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == anode.c.id).join(
        ChromatogramToChromatogramTreeNode,
        ChromatogramToChromatogramTreeNode.c.node_id == anode.c.id)

    branch_peaks_join = peak_join.join(
        anode,
        ChromatogramTreeNodeToDeconvolutedPeak.c.node_id == anode.c.id).join(
        ChromatogramTreeNodeBranch,
        ChromatogramTreeNodeBranch.child_id == anode.c.id).join(
        bnode, ChromatogramTreeNodeBranch.parent_id == bnode.c.id).join(
        ChromatogramToChromatogramTreeNode,
        ChromatogramToChromatogramTreeNode.c.node_id == bnode.c.id)

    branch_intensities = select([apeak.c.intensity, anode.c.retention_time]).where(
        ChromatogramToChromatogramTreeNode.c.chromatogram_id == bindparam("id")
    ).select_from(branch_peaks_join)

    root_intensities = select([apeak.c.intensity, anode.c.retention_time]).where(
        ChromatogramToChromatogramTreeNode.c.chromatogram_id == bindparam("id")
    ).select_from(root_peaks_join)

    all_intensities_q = root_intensities.union_all(branch_intensities).order_by(
        anode.c.retention_time)

    return all_intensities_q


_as_arrays_stmt = __build_as_arrays_query()



from glycresoft.database.builder.glycan import CombinatorialGlycanHypothesisSerializer
from glycresoft.database.prebuilt.utils import hypothesis_register, BuildBase
from io import StringIO

combinatorial_source = u'''HexA  1   9
HexN    1   9
enHexA  0   1
@acetyl 0   9
@sulfate    0   48

@acetyl <=  HexN
@sulfate <= ((HexN * 3 - @acetyl) + (HexA * 2) + (enHexA * 2))
HexA <= HexN + 1
HexN <= HexA + 1
'''

hypothesis_metadata = {
    "name": 'Low Molecular Weight Heparins',
    "hypothesis_type": 'glycan_composition',
    "description": 'A database of heparin and heparan sulfate chains covering'
    ' the combinatorial space of possible saccharides of up to 18 monosaccharides.'
}


@hypothesis_register(hypothesis_metadata['name'])
class LowMolecularWeightHeparansBuilder(BuildBase):

    def get_hypothesis_metadata(self):
        return hypothesis_metadata

    def build(self, database_connection, **kwargs):
        kwargs.setdefault('hypothesis_name', self.hypothesis_metadata['name'])
        task = CombinatorialGlycanHypothesisSerializer(
            StringIO(combinatorial_source), database_connection,
            **kwargs)
        task.start()
        return task



from typing import (Any, Callable, Collection, DefaultDict, Dict, Hashable,
                    List, Optional, Set, Tuple, NamedTuple, Type, Union,
                    TYPE_CHECKING)

from glycresoft.chromatogram_tree import Chromatogram, ChromatogramFilter
from glycresoft.chromatogram_tree.mass_shift import Unmodified

from glycresoft.task import TaskBase

from .base import Predicate, SolutionEntry, TargetType, default_threshold, DEBUG_MODE
from .chromatogram import TandemAnnotatedChromatogram
from .revision import MS2RevisionValidator
from .graph import build_glycopeptide_key, MassShiftDeconvolutionGraph, RepresenterDeconvolution


class GraphAnnotatedChromatogramAggregator(TaskBase):
    annotated_chromatograms: List[TandemAnnotatedChromatogram]
    delta_rt: float
    require_unmodified: bool
    threshold_fn: Predicate
    key_fn: Callable[[TargetType], Hashable]

    revision_validator: MS2RevisionValidator

    def __init__(self, annotated_chromatograms, delta_rt=0.25, require_unmodified=True,
                 threshold_fn=default_threshold, key_fn=build_glycopeptide_key,
                 revision_validator: Optional[MS2RevisionValidator] = None):
        if revision_validator is None:
            revision_validator = MS2RevisionValidator(threshold_fn)
        self.annotated_chromatograms = sorted(annotated_chromatograms,
                                              key=lambda x: (
                                                  max(sset.best_solution(
                                                  ).score for sset in x.tandem_solutions)
                                                  if x.tandem_solutions else -float('inf'), x.total_signal),
                                              reverse=True)
        self.delta_rt = delta_rt
        self.require_unmodified = require_unmodified
        self.threshold_fn = threshold_fn
        self.key_fn = key_fn
        self.revision_validator = revision_validator

    def build_graph(self) -> MassShiftDeconvolutionGraph:
        self.log("Constructing chromatogram graph")
        assignable = []
        rest: List[Chromatogram] = []
        for chrom in self.annotated_chromatograms:
            if chrom.composition is not None:
                assignable.append(chrom)
            else:
                rest.append(chrom)
        graph = MassShiftDeconvolutionGraph(assignable)
        graph.build(self.delta_rt, self.threshold_fn)
        return graph, rest

    def deconvolve(self, graph: MassShiftDeconvolutionGraph):
        components = graph.connected_components()
        assigned = []
        if DEBUG_MODE:
            breakpoint()

        for group in components:
            spectra_in = set()
            for node in group:
                for sset in node.tandem_solutions:
                    spectra_in.add(sset.scan.id)

            deconv = RepresenterDeconvolution(
                group,
                threshold_fn=self.threshold_fn,
                key_fn=self.key_fn,
                revision_validator=self.revision_validator,
            )
            deconv.solve()
            reps = deconv.assign_representers()
            spectra_out = set()
            for node in reps:
                for sset in node.tandem_solutions:
                    spectra_out.add(sset.scan.id)
            if spectra_out != spectra_in:
                self.error(
                    f"{len(spectra_in - spectra_out)} spectra lost while solving this component")
            if DEBUG_MODE and spectra_out != spectra_in:
                breakpoint()
            assigned.extend(reps)
        return assigned

    def run(self):
        graph, rest = self.build_graph()
        solutions = self.deconvolve(graph)
        solutions.extend(rest)
        return ChromatogramFilter(solutions)


class AnnotatedChromatogramAggregator(TaskBase):
    """A basic chromatogram merger that groups chromatograms according to their best analyte.

    This algorithm looks for RT proximity.
    """

    annotated_chromatograms: ChromatogramFilter
    delta_rt: float
    require_unmodified: bool
    threshold_fn: Predicate

    def __init__(self, annotated_chromatograms, delta_rt=0.25, require_unmodified=True,
                 threshold_fn=default_threshold):
        self.annotated_chromatograms = annotated_chromatograms
        self.delta_rt = delta_rt
        self.require_unmodified = require_unmodified
        self.threshold_fn = threshold_fn

    def combine_chromatograms(self, aggregated):
        merged = []
        for entity, group in aggregated.items():
            out = []
            group = sorted(group, key=lambda x: x.start_time)
            chroma = group[0]
            for obs in group[1:]:
                if chroma.chromatogram.overlaps_in_time(obs) or abs(
                        chroma.end_time - obs.start_time) < self.delta_rt:
                    chroma = chroma.merge(obs)
                else:
                    out.append(chroma)
                    chroma = obs
            out.append(chroma)
            merged.extend(out)
        return merged

    def aggregate_by_annotation(self, annotated_chromatograms: ChromatogramFilter) -> Tuple[
                                                                                            List[TandemAnnotatedChromatogram],
                                                                                            DefaultDict[
                                                                                                str,
                                                                                                List[TandemAnnotatedChromatogram]
                                                                                            ]
                                                                                        ]:
        """Group chromatograms by their assigned entity."""
        finished = []
        aggregated = DefaultDict(list)
        for chroma in annotated_chromatograms:
            if chroma.composition is not None:
                if chroma.entity is not None:
                    # Convert to string to avoid redundant sequences from getting
                    # binned differently due to random ordering of ids.
                    aggregated[str(chroma.entity)].append(chroma)
                else:
                    aggregated[str(chroma.composition)].append(chroma)
            else:
                finished.append(chroma)
        return finished, aggregated

    def aggregate(self, annotated_chromatograms: ChromatogramFilter) -> List[TandemAnnotatedChromatogram]:
        """Drives the aggregation process."""
        self.log("Aggregating Common Entities: %d chromatograms" %
                 (len(annotated_chromatograms,)))
        finished, aggregated = self.aggregate_by_annotation(
            annotated_chromatograms)
        finished.extend(self.combine_chromatograms(aggregated))
        self.log("After Merging: %d chromatograms" % (len(finished),))
        return finished

    def replace_solution(self, chromatogram: TandemAnnotatedChromatogram, solutions: List[SolutionEntry]):
        # select the best solution
        solutions = sorted(
            solutions, key=lambda x: x.score, reverse=True)

        # remove the invalidated mass shifts so that we're dealing with the raw masses again.
        current_shifts = chromatogram.chromatogram.mass_shifts
        if len(current_shifts) > 1:
            self.log("... Found a multiply shifted identification with no Unmodified state: %s" % (
                chromatogram.entity, ))
        partitions = []
        for shift in current_shifts:
            # The _ collects the "not modified with shift" portion of the chromatogram
            # we'll strip out in successive iterations. By virtue of reaching this point,
            # we're never dealing with an Unmodified portion.
            partition, _ = chromatogram.chromatogram.bisect_mass_shift(shift)
            # If we're somehow dealing with a compound mass shift here, remove
            # the bad shift from the composite, and reset the modified nodes to
            # Unmodified.
            partitions.append(partition.deduct_node_type(shift))

        # Merge in
        accumulated_chromatogram = partitions[0]
        for partition in partitions[1:]:
            accumulated_chromatogram = accumulated_chromatogram.merge(
                partition)
        chromatogram.chromatogram = accumulated_chromatogram

        # update the tandem annotations
        for solution_set in chromatogram.tandem_solutions:
            solution_set.mark_top_solutions(reject_shifted=self.require_unmodified)
        chromatogram.assign_entity(
            solutions[0],
            entity_chromatogram_type=chromatogram.chromatogram.__class__)
        chromatogram.representative_solutions = solutions
        return chromatogram

    def reassign_modified_only_cases(self, annotated_chromatograms):
        out = []
        for chromatogram in annotated_chromatograms:
            # the structure's best match has not been identified in an unmodified state
            if Unmodified not in chromatogram.mass_shifts:
                original_entity = getattr(chromatogram, "entity")
                solutions = chromatogram.most_representative_solutions(
                    self.threshold_fn, reject_shifted=True)
                # if there is a reasonable solution in an unmodified state
                if solutions:
                    chromatogram = self.replace_solution(
                        chromatogram, solutions)
                    self.debug("... Replacing %s with %s",
                               original_entity, chromatogram.entity)
                    out.append(chromatogram)
                else:
                    self.log(
                        "... Could not find an alternative option for %r" % (chromatogram,))
                    out.append(chromatogram)
            else:
                out.append(chromatogram)
        return out

    def run(self):
        merged_chromatograms = self.aggregate(self.annotated_chromatograms)
        if self.require_unmodified:
            spliced = self.reassign_modified_only_cases(merged_chromatograms)
            merged_chromatograms = self.aggregate(spliced)
        result = ChromatogramFilter(merged_chromatograms)
        return result


def aggregate_by_assigned_entity(annotated_chromatograms: List[TandemAnnotatedChromatogram],
                                 delta_rt: float = 0.25,
                                 require_unmodified: bool = True,
                                 threshold_fn: Predicate = default_threshold,
                                 revision_validator: Optional[MS2RevisionValidator] = None):
    job = GraphAnnotatedChromatogramAggregator(
        annotated_chromatograms,
        delta_rt=delta_rt,
        require_unmodified=require_unmodified,
        threshold_fn=threshold_fn,
        revision_validator=revision_validator
    )
    finished = job.run()
    return finished



import csv
import logging

from io import TextIOWrapper
from typing import TYPE_CHECKING, Dict, Iterable, List, Mapping, TextIO

from six import PY2

from glycresoft.task import TaskBase

from glycresoft.scoring.elution_time_grouping import (
    GlycopeptideChromatogramProxy
)

if TYPE_CHECKING:
    from glycresoft.serialize import Analysis

status_logger = logging.getLogger("glycresoft.status")


def csv_stream(outstream):
    if 'b' in outstream.mode:
        if not PY2:
            return TextIOWrapper(outstream, 'utf8', newline="")
        else:
            return outstream
    else:
        import warnings
        warnings.warn("Opened CSV stream in text mode!")
        return outstream


class CSVSerializerBase(TaskBase):
    entity_label: str = "entities"
    log_interval: int = 100
    outstream: TextIO
    writer: csv.writer
    delimiter: str
    _entities_iterable: Iterable

    def __init__(self, outstream, entities_iterable, delimiter=','):
        self.outstream = outstream
        try:
            self.is_binary = 'b' in self.outstream.mode
        except AttributeError:
            self.is_binary = True
        if self.is_binary:
            try:
                self.outstream = TextIOWrapper(outstream, 'utf8', newline="")
            except AttributeError:
                # must be Py2
                pass
        self.writer = csv.writer(self.outstream, delimiter=delimiter)
        self._entities_iterable = entities_iterable

    def status_update(self, message: str):
        status_logger.info(message)

    def writerows(self, iterable: Iterable[Iterable[str]]):
        self.writer.writerows(iterable)

    def writerow(self, row: Iterable[str]):
        self.writer.writerow(row)

    def get_header(self) -> List[str]:
        raise NotImplementedError()

    def convert_object(self, obj):
        raise NotImplementedError()

    @property
    def header(self):
        return self.get_header()

    def run(self):
        if self.header:
            header = self.header
            self.writerow(header)
        entity_label = self.entity_label
        log_interval = self.log_interval
        gen = (self.convert_object(entity) for entity in self._entities_iterable)
        i = 0
        for i, row in enumerate(gen, 1):
            if i % log_interval == 0 and i != 0:
                self.status_update(f"Handled {i} {entity_label}")
            self.writerow(row)
        self.status_update(f"Handled {i} {entity_label}")
        self.outstream.flush() # If the TextIOWrapper is buffering, we need to flush it here


class GlycanHypothesisCSVSerializer(CSVSerializerBase):
    def __init__(self, outstream, entities_iterable, delimiter=','):
        super(GlycanHypothesisCSVSerializer, self).__init__(
            outstream, entities_iterable, delimiter)

    def get_header(self):
        return [
            "composition",
            "calculated_mass",
            "classes"
        ]

    def convert_object(self, obj):
        attribs = [
            obj.composition,
            obj.calculated_mass,
            ";".join([sc.name for sc in obj.structure_classes])
        ]
        return list(map(str, attribs))


class ImportableGlycanHypothesisCSVSerializer(CSVSerializerBase):
    def __init__(self, outstream, entities_iterable):
        super(ImportableGlycanHypothesisCSVSerializer, self).__init__(
            outstream, entities_iterable, delimiter="\t")

    def get_header(self):
        return False

    def convert_object(self, obj):
        attribs = [
            obj.composition,
        # Use list concatenation to ensure that the glycan classes
        # are treated as independent entries not containing whitespace
        # so that they are not quoted
        ] + [sc.name for sc in obj.structure_classes]
        return list(map(str, attribs))


class GlycopeptideHypothesisCSVSerializer(CSVSerializerBase):
    def __init__(self, outstream, entities_iterable, delimiter=','):
        super(GlycopeptideHypothesisCSVSerializer, self).__init__(
            outstream, entities_iterable, delimiter)

    def get_header(self):
        return [
            "glycopeptide",
            "calculated_mass",
            "start_position",
            "end_position",
            "protein",
        ]

    def convert_object(self, obj):
        attribs = [
            obj.glycopeptide_sequence,
            obj.calculated_mass,
            obj.peptide.start_position,
            obj.peptide.end_position,
            obj.peptide.protein.name
        ]
        return list(map(str, attribs))


class SimpleChromatogramCSVSerializer(CSVSerializerBase):
    def __init__(self, outstream, entities_iterable, delimiter=','):
        super(SimpleChromatogramCSVSerializer, self).__init__(
            outstream, entities_iterable, delimiter)

    def get_header(self):
        return [
            "neutral_mass",
            "total_signal",
            "charge_states",
            "start_time",
            "apex_time",
            "end_time"
        ]

    def convert_object(self, obj):
        attribs = [
            obj.weighted_neutral_mass,
            obj.total_signal,
            ';'.join(map(str, obj.charge_states)),
            obj.start_time,
            obj.apex_time,
            obj.end_time,
        ]
        return map(str, attribs)


class SimpleScoredChromatogramCSVSerializer(SimpleChromatogramCSVSerializer):

    def get_header(self):
        headers = super(SimpleScoredChromatogramCSVSerializer, self).get_header()
        headers += [
            "score",
            "line_score",
            "isotopic_fit",
            "spacing_fit",
            "charge_count",
        ]
        return headers

    def convert_object(self, obj):
        attribs = super(SimpleScoredChromatogramCSVSerializer, self).convert_object(obj)
        scores = obj.score_components()
        more_attribs = [
            obj.score,
            scores.line_score,
            scores.isotopic_fit,
            scores.spacing_fit,
            scores.charge_count,
        ]
        attribs = list(attribs) + list(map(str, more_attribs))
        return attribs


class GlycanLCMSAnalysisCSVSerializer(CSVSerializerBase):
    def __init__(self, outstream, entities_iterable, delimiter=','):
        super(GlycanLCMSAnalysisCSVSerializer, self).__init__(
            outstream, entities_iterable, delimiter)

    def get_header(self):
        return [
            "composition",
            "neutral_mass",
            "mass_accuracy",
            "score",
            "total_signal",
            "start_time",
            "end_time",
            "apex_time",
            "charge_states",
            "mass_shifts",
            "line_score",
            "isotopic_fit",
            "spacing_fit",
            "charge_count",
            "ambiguous_with",
            "used_as_mass_shift"
        ]

    def convert_object(self, obj):
        scores = obj.score_components()
        attribs = [
            obj.glycan_composition,
            obj.weighted_neutral_mass,
            0 if obj.glycan_composition is None else (
                (obj.glycan_composition.mass() - obj.weighted_neutral_mass
                 ) / obj.weighted_neutral_mass),
            obj.score,
            obj.total_signal,
            obj.start_time,
            obj.end_time,
            obj.apex_time,
            ';'.join(map(str, obj.charge_states)),
            ';'.join([a.name for a in obj.mass_shifts]),
            scores.line_score,
            scores.isotopic_fit,
            scores.spacing_fit,
            scores.charge_count,
            ';'.join("%s:%s" % (p[0], p[1].name) for p in obj.ambiguous_with),
            ';'.join("%s:%s" % (p[0], p[1].name) for p in obj.used_as_mass_shift)
        ]
        return list(map(str, attribs))


class GlycopeptideLCMSMSAnalysisCSVSerializer(CSVSerializerBase):
    analysis: 'Analysis'
    protein_name_resolver: Mapping[int, str]
    include_group: bool

    def __init__(self, outstream, entities_iterable, protein_name_resolver, analysis, delimiter=',',
                 include_group: bool=True):
        super(GlycopeptideLCMSMSAnalysisCSVSerializer, self).__init__(outstream, entities_iterable, delimiter)
        self.protein_name_resolver = protein_name_resolver
        self.analysis = analysis
        self.retention_time_model = analysis.parameters.get("retention_time_model")
        self.include_group = include_group

    def get_header(self):
        headers = [
            "glycopeptide",
            "analysis",
            "neutral_mass",
            "mass_accuracy",
            "ms1_score",
            "ms2_score",
            "q_value",
            "total_signal",
            "start_time",
            "end_time",
            "apex_time",
            "charge_states",
            "msms_count",
            "peptide_start",
            "peptide_end",
            "protein_name",
            "n_glycosylation_sites",
            "mass_shifts",
        ]
        if self.include_group:
            headers.append("group_id")
        if self.retention_time_model:
            headers.extend([
                "predicted_apex_interval_start",
                "predicted_apex_interval_end",
                "retention_time_score",
            ])
        return headers

    def convert_object(self, obj):
        has_chrom = obj.has_chromatogram()
        if has_chrom:
            weighted_neutral_mass = obj.weighted_neutral_mass
            charge_states = obj.charge_states
        else:
            weighted_neutral_mass = obj.tandem_solutions[0].scan.precursor_information.neutral_mass
            charge_states = (obj.tandem_solutions[0].scan.precursor_information.charge,)

        attribs = [
            str(obj.structure),
            self.analysis.name,
            weighted_neutral_mass,
            (obj.structure.total_mass - weighted_neutral_mass) / weighted_neutral_mass,
            obj.ms1_score,
            obj.ms2_score,
            obj.q_value,
            obj.total_signal,
            obj.start_time if has_chrom else '',
            obj.end_time if has_chrom else '',
            obj.apex_time if has_chrom else '',
            ";".join(map(str, charge_states)),
            len(obj.spectrum_matches),
            obj.protein_relation.start_position,
            obj.protein_relation.end_position,
            self.protein_name_resolver[obj.protein_relation.protein_id],
            ';'.join([
                str(i + obj.start_position) for i in obj.structure.n_glycan_sequon_sites
            ]),
            ';'.join([a.name for a in obj.mass_shifts]),
        ]
        try:
            attribs.append(obj.ambiguous_id)
        except AttributeError:
            attribs.append(-1)
        if self.retention_time_model:
            if has_chrom:
                proxy = GlycopeptideChromatogramProxy.from_chromatogram(obj)
                rt_start, rt_end = self.retention_time_model.predict_interval(proxy, 0.01)
                rt_score = self.retention_time_model.score_interval(proxy, 0.01)
                attribs.append(rt_start)
                attribs.append(rt_end)
                attribs.append(rt_score)
            else:
                attribs.append("-")
                attribs.append("-")
                attribs.append("-")
        return list(map(str, attribs))


class MultiScoreGlycopeptideLCMSMSAnalysisCSVSerializer(GlycopeptideLCMSMSAnalysisCSVSerializer):
    def get_header(self):
        headers = super(MultiScoreGlycopeptideLCMSMSAnalysisCSVSerializer, self).get_header()
        headers.extend([
            "glycopeptide_score",
            "peptide_score",
            "glycan_score",
            "glycan_coverage",
            "total_q_value",
            "peptide_q_value",
            "glycan_q_value",
            "glycopeptide_q_value",
            "localizations"
        ])
        return headers

    def convert_object(self, obj):
        fields = super(MultiScoreGlycopeptideLCMSMSAnalysisCSVSerializer, self).convert_object(obj)
        score_set = obj.score_set
        q_value_set = obj.q_value_set
        new_fields = [
            score_set.glycopeptide_score if score_set is not None else "?",
            score_set.peptide_score if score_set is not None else "?",
            score_set.glycan_score if score_set is not None else "?",
            score_set.glycan_coverage if score_set is not None else "?",
            q_value_set.total_q_value if q_value_set is not None else "?",
            q_value_set.peptide_q_value if q_value_set is not None else "?",
            q_value_set.glycan_q_value if q_value_set is not None else "?",
            q_value_set.glycopeptide_q_value if q_value_set is not None else "?",
            ';'.join(map(str, obj.localizations))
        ]
        fields.extend(map(str, new_fields))
        return fields


class GlycopeptideSpectrumMatchAnalysisCSVSerializer(CSVSerializerBase):
    entity_label = "glycopeptide spectrum matches"
    log_interval = 5000

    include_rank: bool
    include_group: bool
    analysis: 'Analysis'
    protein_name_resolver: Dict[int, str]

    def __init__(self, outstream, entities_iterable, protein_name_resolver, analysis, delimiter=',',
                 include_rank: bool = True, include_group: bool = True):
        super(GlycopeptideSpectrumMatchAnalysisCSVSerializer, self).__init__(outstream, entities_iterable, delimiter)
        self.protein_name_resolver = protein_name_resolver
        self.analysis = analysis
        self.include_rank = include_rank
        self.include_group = include_group

    def get_header(self):
        names = [
            "glycopeptide",
            "analysis",
            "neutral_mass",
            "mass_accuracy",
            "mass_shift_name",
            "scan_id",
            "scan_time",
            "charge",
            "ms2_score",
            "q_value",
            "precursor_abundance",
            "peptide_start",
            "peptide_end",
            "protein_name",
            "n_glycosylation_sites",
            "is_best_match",
            "is_precursor_fit",
        ]
        if self.include_rank:
            names.append("rank")
        if self.include_group:
            names.append("group_id")
        return names

    def convert_object(self, obj):
        try:
            mass_shift = obj.mass_shift
            mass_shift_mass = mass_shift.mass
            mass_shift_name = mass_shift.name
        except Exception:
            mass_shift_name = "Unmodified"
            mass_shift_mass = 0
        target = obj.target
        scan = obj.scan
        precursor = scan.precursor_information
        precursor_mass = precursor.neutral_mass
        attribs = [
            str(obj.target),
            self.analysis.name,
            precursor_mass,
            ((target.total_mass + mass_shift_mass) -
             precursor_mass) / precursor_mass,
            mass_shift_name,
            scan.scan_id,
            scan.scan_time,
            scan.precursor_information.extracted_charge,
            obj.score,
            obj.q_value,
            scan.precursor_information.intensity,
            target.protein_relation.start_position,
            target.protein_relation.end_position,
            self.protein_name_resolver[target.protein_relation.protein_id],
            ';'.join([
                str(i + obj.target.protein_relation.start_position)
                for i in obj.target.n_glycan_sequon_sites
            ]),
            obj.is_best_match,
            not precursor.defaulted
        ]
        if self.include_rank:
            try:
                rank = obj.rank
            except AttributeError:
                rank = -1
            attribs.append(rank)
        if self.include_group:
            try:
                group = obj.cluster_id
            except AttributeError:
                group = -1
            attribs.append(group)
        return list(map(str, attribs))


class MultiScoreGlycopeptideSpectrumMatchAnalysisCSVSerializer(GlycopeptideSpectrumMatchAnalysisCSVSerializer):

    def get_header(self):
        header = super(MultiScoreGlycopeptideSpectrumMatchAnalysisCSVSerializer, self).get_header()
        header.extend([
            "peptide_score",
            "glycan_score",
            "glycan_coverage",
            "peptide_q_value",
            "glycan_q_value",
            "glycopeptide_q_value",
            "localizations",
        ])
        return header

    def convert_object(self, obj):
        fields = super(MultiScoreGlycopeptideSpectrumMatchAnalysisCSVSerializer, self).convert_object(obj)
        score_set = obj.score_set
        fdr_set = obj.q_value_set
        fields.extend([
            score_set.peptide_score,
            score_set.glycan_score,
            score_set.glycan_coverage,
            fdr_set.peptide_q_value,
            fdr_set.glycan_q_value,
            fdr_set.glycopeptide_q_value,
            ';'.join(map(str, obj.localizations))
        ])
        return fields



'''Extra logic for fitting a model across multiple LC-MS runs.
'''
from collections import defaultdict

import numpy as np

try:
    from matplotlib import pyplot as plt
except ImportError:
    pass


import glycopeptidepy

from .model import AbundanceWeightedPeptideFactorElutionTimeFitter, make_counter
from .structure import CommonGlycopeptide
from .linear_regression import weighted_linear_regression_fit


class ReplicatedAbundanceWeightedPeptideFactorElutionTimeFitter(AbundanceWeightedPeptideFactorElutionTimeFitter):
    def __init__(self, chromatograms, factors=None, scale=1, replicate_key_attr='analysis_name',
                 use_retention_time_normalization=False, reference_sample=None):
        if replicate_key_attr is None:
            replicate_key_attr = 'analysis_name'
        if factors is None:
            factors = ['Hex', 'HexNAc', 'Fuc', 'Neu5Ac']
        self.replicate_key_attr = replicate_key_attr
        self.reference_sample = reference_sample
        self._replicate_to_indicator = defaultdict(make_counter(0))
        self.use_retention_time_normalization = use_retention_time_normalization
        self.run_normalizer = None
        # Ensure that _replicate_to_indicator is properly initialized
        # for obs in chromatograms:
        #     _ = self._replicate_to_indicator[self._get_replicate_key(obs)]
        self.chromatograms = chromatograms

        index, samples = self._build_common_glycopeptides()
        self.replicate_names = samples
        if self.use_retention_time_normalization:
            self.run_normalizer = LinearRetentionTimeCorrector(index, samples, self.reference_sample)
            if self.reference_sample is None:
                self.reference_sample = self.run_normalizer.reference_key
            self.run_normalizer.fit()
        else:
            if self.reference_sample is None:
                self.reference_sample = samples[0]
        _ = self._replicate_to_indicator[self.reference_sample]
        for sample_key in samples:
            _ = self._replicate_to_indicator[sample_key]
        super(ReplicatedAbundanceWeightedPeptideFactorElutionTimeFitter, self).__init__(
            chromatograms, list(factors), scale)

    def reset_reference_sample(self, reference_sample):
        self.reference_sample = reference_sample
        self._replicate_to_indicator = defaultdict(make_counter(0))
        _ = self._replicate_to_indicator[self.reference_sample]
        for sample_key in self.replicate_names:
            _ = self._replicate_to_indicator[sample_key]
        self.data = self._prepare_data_matrix(self.neutral_mass_array)

    def _get_apex_time(self, chromatogram):
        value = super(ReplicatedAbundanceWeightedPeptideFactorElutionTimeFitter,
                      self)._get_apex_time(chromatogram)
        if self.use_retention_time_normalization and self.run_normalizer is not None:
            replicate_id = self._get_replicate_key(chromatogram)
            value = self.run_normalizer.correct(value, replicate_id)
        return value

    def _build_common_glycopeptides(self):
        index = defaultdict(dict)
        for case in self.chromatograms:
            index[str(case.structure)][self._get_replicate_key(case)] = case
        result = [CommonGlycopeptide(glycopeptidepy.parse(k), v) for k, v in index.items()]
        samples = set()
        for case in result:
            samples.update(case.keys())
        return result, sorted(samples)

    def _get_replicate_key(self, chromatogram):
        if isinstance(self.replicate_key_attr, (str, unicode)):
            return getattr(chromatogram, self.replicate_key_attr)
        elif callable(self.replicate_key_attr):
            return self.replicate_key_attr(chromatogram)

    def _prepare_data_matrix(self, mass_array):
        design_matrix = super(
            ReplicatedAbundanceWeightedPeptideFactorElutionTimeFitter,
            self)._prepare_data_matrix(mass_array)
        p = len(self._replicate_to_indicator)
        n = len(self.chromatograms)
        replicate_matrix = np.zeros((p, n))
        indicator = dict(self._replicate_to_indicator)
        for i, c in enumerate(self.chromatograms):
            try:
                # Here, if one of the levels is not omitted, the matrix will be linearly dependent.
                # So drop the 0th factor level.
                j = indicator[self._get_replicate_key(c)]
                if j != 0:
                    replicate_matrix[j, i] = 1
            except KeyError:
                pass
        return np.hstack((replicate_matrix.T, design_matrix))

    def feature_names(self):
        names = []
        replicates = [None] * len(self._replicate_to_indicator)
        for key, value in self._replicate_to_indicator.items():
            replicates[value] = key
        names.extend(replicates)
        peptides = [None] * len(self._peptide_to_indicator)
        for key, value in self._peptide_to_indicator.items():
            peptides[value] = key
        names.extend(peptides)
        names.extend(self.factors)
        return names

    def _prepare_data_vector(self, chromatogram, no_intercept=False):
        data_vector = super(
            ReplicatedAbundanceWeightedPeptideFactorElutionTimeFitter,
            self)._prepare_data_vector(chromatogram, no_intercept=no_intercept)
        p = len(self._replicate_to_indicator)
        replicates = [0 for _ in range(p)]
        indicator = dict(self._replicate_to_indicator)
        if not no_intercept:
            try:
                # Here, if one of the levels is not omitted, the matrix will be linearly dependent.
                # So drop the 0th factor level.
                j = self._get_replicate_key(chromatogram)
                if j != 0:
                    replicates[indicator[j]] = 1
            except KeyError:
                import warnings
                warnings.warn(
                    "Replicate Key %s not part of the model." % (j, ))
        return np.hstack((replicates, data_vector))


class LinearRetentionTimeCorrector(object):
    def __init__(self, common_observations, sample_keys, reference_key=None):
        self.common_observations = common_observations
        self.sample_keys = sample_keys
        self.reference_key = reference_key
        self.sample_to_correction_parameters = dict()
        self.sample_to_vector = dict()
        if self.reference_key is None:
            self.reference_key = self.select_reference()

    def build_sample_distance_matrix(self):
        n = len(self.sample_keys)
        distance_matrix = np.zeros((n, n))
        for i in range(n):
            ivec = self._make_vector_for(self.sample_keys[i])
            for j in range(i + 1, n):
                jvec = self._make_vector_for(self.sample_keys[j])
                wls = self._fit_correction(jvec, ivec)
                distance = wls.R2
                distance_matrix[i, j] = distance
                distance_matrix[j, i] = distance
        return distance_matrix

    def select_reference(self):
        distance_matrix = self.build_sample_distance_matrix()
        return self.sample_keys[np.argmin(distance_matrix.sum(axis=0))]

    def _make_vector_for(self, sample):
        if sample in self.sample_to_vector:
            return self.sample_to_vector[sample]
        result = []
        for obs in self.common_observations:
            try:
                result.append(obs[sample].apex_time)
            except KeyError:
                result.append(np.nan)
        result = self.sample_to_vector[sample] = np.array(result)
        return result

    def _fit_correction(self, a, b):
        delta = b - a
        mask = ~(np.isnan(delta))
        wls_fit = weighted_linear_regression_fit(
            b[mask], delta[mask], prepare=True)
        return wls_fit

    def fit_correction_for(self, sample_key):
        sample = self._make_vector_for(sample_key)
        reference = self._make_vector_for(self.reference_key)
        wls_fit = self._fit_correction(sample, reference)
        self.sample_to_correction_parameters[sample_key] = wls_fit
        return wls_fit

    def fit(self):
        for sample_key in self.sample_keys:
            if sample_key == self.reference_key:
                continue
            self.fit_correction_for(sample_key)

    def correct(self, x, sample_key):
        return x - self._correction_factor_for(x, sample_key)

    def _correction_factor_for(self, x, sample_key):
        if sample_key == self.reference_key:
            return 0
        fit = self.sample_to_correction_parameters[sample_key]
        return (fit.parameters[1] * x + fit.parameters[0])

    def plot(self, ax=None):
        if ax is None:
            _fig, ax = plt.subplots(1)
        X = self._make_vector_for(self.reference_key)
        mask = ~np.isnan(X)
        X = X[mask]
        Xhat = np.arange(X.min(), X.max())
        for sample_key in sorted(self.sample_keys):
            if sample_key == self.reference_key:
                ax.scatter(X, np.zeros_like(X), label="%s Reference" % (sample_key, ))
            else:
                fit = self.sample_to_correction_parameters[sample_key]
                ax.scatter(fit.data[0][:, 1], fit.data[1],
                        label=r'%s %0.2fX + %0.2f ($R^2$=%0.2f)' % (
                            sample_key, fit.parameters[1], fit.parameters[0], fit.R2))
                Yhat = Xhat * fit.parameters[1] + fit.parameters[0]
                ax.plot(Xhat, Yhat, linestyle='--')
        legend = ax.legend(fontsize=6)
        frame = legend.get_frame()
        frame.set_linewidth(0.5)
        frame.set_alpha(0.5)
        ax.set_xlabel("Time")
        ax.set_ylabel(r"$\delta$ Time")
        return ax




from typing import Optional

from glycopeptidepy.structure.residue import UnknownAminoAcidException

from glycresoft.serialize import func
from glycresoft.serialize.hypothesis.peptide import (
    Peptide, Protein, Glycopeptide)
from glycresoft.serialize.utils import toggle_indices

from six import string_types as basestring

from .proteomics.peptide_permutation import (
    ProteinDigestor,
    MultipleProcessProteinDigestor,
    UniprotProteinAnnotator)

from .proteomics.fasta import open_fasta
from .common import (
    GlycopeptideHypothesisSerializerBase,
    PeptideGlycosylator, PeptideGlycosylatingProcess,
    NonSavingPeptideGlycosylatingProcess,
    MultipleProcessPeptideGlycosylator, ProteinReversingMixin)




class FastaGlycopeptideHypothesisSerializer(GlycopeptideHypothesisSerializerBase):
    def __init__(self, fasta_file, connection, glycan_hypothesis_id, hypothesis_name=None,
                 protease='trypsin', constant_modifications=None, variable_modifications=None,
                 max_missed_cleavages=2, max_glycosylation_events=1, semispecific=False,
                 max_variable_modifications=None, full_cross_product=True,
                 peptide_length_range=(5, 60), require_glycosylation_sites=True,
                 use_uniprot=True, include_cysteine_n_glycosylation: bool=False,
                 uniprot_source_file: Optional[str]=None):
        GlycopeptideHypothesisSerializerBase.__init__(
            self, connection, hypothesis_name, glycan_hypothesis_id, full_cross_product,
            use_uniprot=use_uniprot,
            uniprot_source_file=uniprot_source_file)
        self.fasta_file = fasta_file
        self.protease = protease
        self.constant_modifications = constant_modifications
        self.variable_modifications = variable_modifications
        self.max_missed_cleavages = max_missed_cleavages
        self.max_glycosylation_events = max_glycosylation_events
        self.semispecific = semispecific
        self.max_variable_modifications = max_variable_modifications
        self.peptide_length_range = peptide_length_range or (5, 60)
        self.require_glycosylation_sites = require_glycosylation_sites
        self.include_cysteine_n_glycosylation = include_cysteine_n_glycosylation

        params = {
            "fasta_file": fasta_file,
            "enzymes": [protease] if isinstance(protease, basestring) else list(protease),
            "constant_modifications": constant_modifications,
            "variable_modifications": variable_modifications,
            "max_missed_cleavages": max_missed_cleavages,
            "max_glycosylation_events": max_glycosylation_events,
            "semispecific": semispecific,
            "max_variable_modifications": max_variable_modifications,
            "full_cross_product": self.full_cross_product,
            "peptide_length_range": self.peptide_length_range,
            "require_glycosylation_sites": self.require_glycosylation_sites,
            "include_cysteine_n_glycosylation": self.include_cysteine_n_glycosylation,
            "uniprot_source_file": self.uniprot_source_file,
        }
        self.set_parameters(params)

    def extract_proteins(self):
        i = 0
        protein: Protein
        for protein in open_fasta(self.fasta_file):
            protein.hypothesis_id = self.hypothesis_id
            protein._init_sites(
                include_cysteine_n_glycosylation=self.include_cysteine_n_glycosylation)

            self.session.add(protein)
            i += 1
            if i % 5000 == 0:
                self.log("... %d Proteins Extracted" % (i,))
                self.session.commit()

        self.session.commit()
        self.log("%d Proteins Extracted" % (i,))
        return i

    def protein_ids(self):
        return [i[0] for i in self.query(Protein.id).filter(Protein.hypothesis_id == self.hypothesis_id).all()]

    def _make_digestor(self):
        digestor = ProteinDigestor(
            self.protease, self.constant_modifications, self.variable_modifications,
            self.max_missed_cleavages,
            min_length=self.peptide_length_range[0],
            max_length=self.peptide_length_range[1],
            semispecific=self.semispecific,
            require_glycosylation_sites=self.require_glycosylation_sites,
            include_cysteine_n_glycosylation=self.include_cysteine_n_glycosylation)
        return digestor

    def digest_proteins(self):
        digestor = self._make_digestor()
        i = 0
        j = 0
        protein_ids = self.protein_ids()
        n = len(protein_ids)
        interval = min(n / 10., 100000)
        acc = []
        for protein_id in protein_ids:
            i += 1
            protein = self.query(Protein).get(protein_id)
            if i % interval == 0:
                self.log("... %0.3f%% Complete (%d/%d). %d Peptides Produced." % (i * 100. / n, i, n, j))
            for peptide in digestor.process_protein(protein):
                acc.append(peptide)
                j += 1
                if len(acc) > 100000:
                    self.session.bulk_save_objects(acc)
                    self.session.commit()
                    acc = []
        self.session.bulk_save_objects(acc)
        self.session.commit()
        acc = []

    def split_proteins(self):
        annotator = UniprotProteinAnnotator(
            self,
            self.protein_ids(),
            self.constant_modifications,
            self.variable_modifications,
            include_cysteine_n_glycosylation=self.include_cysteine_n_glycosylation,
            uniprot_source_file=self.uniprot_source_file)
        annotator.run()
        self.deduplicate_peptides()

    def glycosylate_peptides(self):
        glycosylator = PeptideGlycosylator(self.session, self.hypothesis_id)
        acc = []
        i = 0
        for peptide_id in self.peptide_ids():
            peptide = self.query(Peptide).get(peptide_id)
            for glycopeptide in glycosylator.handle_peptide(peptide):
                acc.append(glycopeptide)
                i += 1
                if len(acc) > 100000:
                    self.session.bulk_insert_mappings(Glycopeptide, acc, render_nulls=True)
                    self.session.commit()
                    acc = []
        self.session.bulk_insert_mappings(Glycopeptide, acc, render_nulls=True)
        self.session.commit()

    def run(self):
        self.log("Extracting Proteins")
        self.extract_proteins()
        self.log("Digesting Proteins")
        index_toggler = toggle_indices(self.session, Peptide)
        index_toggler.drop()
        self.digest_proteins()
        self.log("Rebuilding Peptide Index")
        index_toggler.create()
        if self.use_uniprot:
            self.split_proteins()
        self.log("Combinating Glycans")
        self.combinate_glycans(self.max_glycosylation_events)
        if self.full_cross_product:
            self.log("Building Glycopeptides")
            self.glycosylate_peptides()
        self._sql_analyze_database()
        if self.full_cross_product:
            self._count_produced_glycopeptides()
        self.log("Done")


class MultipleProcessFastaGlycopeptideHypothesisSerializer(FastaGlycopeptideHypothesisSerializer):
    def __init__(self, fasta_file, connection, glycan_hypothesis_id, hypothesis_name=None,
                 protease='trypsin', constant_modifications=None, variable_modifications=None,
                 max_missed_cleavages=2, max_glycosylation_events=1, semispecific=False,
                 max_variable_modifications=None, full_cross_product=True, peptide_length_range=(5, 60),
                 require_glycosylation_sites: bool=True, use_uniprot: bool=True,
                 include_cysteine_n_glycosylation: bool=False,
                 uniprot_source_file: Optional[str]=None,
                 n_processes: int=4):
        super(MultipleProcessFastaGlycopeptideHypothesisSerializer, self).__init__(
            fasta_file, connection, glycan_hypothesis_id, hypothesis_name,
            protease=protease,
            constant_modifications=constant_modifications,
            variable_modifications=variable_modifications,
            max_missed_cleavages=max_missed_cleavages,
            max_glycosylation_events=max_glycosylation_events,
            semispecific=semispecific,
            max_variable_modifications=max_variable_modifications,
            full_cross_product=full_cross_product,
            peptide_length_range=peptide_length_range,
            require_glycosylation_sites=require_glycosylation_sites,
            use_uniprot=use_uniprot,
            include_cysteine_n_glycosylation=include_cysteine_n_glycosylation,
            uniprot_source_file=uniprot_source_file)
        self.n_processes = n_processes

    def digest_proteins(self):
        digestor = self._make_digestor()
        task = MultipleProcessProteinDigestor(
            self._original_connection,
            self.hypothesis_id,
            self.protein_ids(),
            digestor, n_processes=self.n_processes)
        task.run()
        n_peptides = self.query(func.count(Peptide.id)).filter(
            Peptide.hypothesis_id == self.hypothesis_id).scalar()
        self.log("%d Base Peptides Produced" % (n_peptides,))

    def _spawn_glycosylator(self, input_queue, done_event):
        return PeptideGlycosylatingProcess(
            self._original_connection, self.hypothesis_id, input_queue,
            chunk_size=3500, done_event=done_event)

    def glycosylate_peptides(self):
        dispatcher = MultipleProcessPeptideGlycosylator(
            self._original_connection, self.hypothesis_id,
            glycan_combination_count=self.total_glycan_combination_count,
            n_processes=self.n_processes)
        dispatcher.process(self.peptide_ids())


_MPFGHS = MultipleProcessFastaGlycopeptideHypothesisSerializer

class ReversingMultipleProcessFastaGlycopeptideHypothesisSerializer(_MPFGHS, ProteinReversingMixin):
    def extract_proteins(self):
        i = 0
        for protein in open_fasta(self.fasta_file):
            try:
                protein = self.reverse_protein(protein)
            except UnknownAminoAcidException:
                continue

            self.session.add(protein)
            i += 1
            if i % 5000 == 0:
                self.log("... %d Proteins Extracted" % (i,))
                self.session.commit()

        self.session.commit()
        self.log("%d Proteins Extracted" % (i,))
        return i


class NonSavingMultipleProcessFastaGlycopeptideHypothesisSerializer(_MPFGHS):
    def _spawn_glycosylator(self, input_queue, done_event):
        return NonSavingPeptideGlycosylatingProcess(
            self._original_connection, self.hypothesis_id, input_queue,
            chunk_size=3500, done_event=done_event)



from .base import (
    ScoringFeatureBase,
    DummyFeature,
    epsilon)

from .shape_fitter import (
    ChromatogramShapeFitter, ChromatogramShapeModel, AdaptiveMultimodalChromatogramShapeFitter)

from .spacing_fitter import (
    ChromatogramSpacingFitter, ChromatogramSpacingModel,
    PartitionAwareRelativeScaleChromatogramSpacingFitter,
    total_intensity)

from .isotopic_fit import (
    IsotopicPatternConsistencyFitter, IsotopicPatternConsistencyModel)

from .charge_state import (
    UniformChargeStateScoringModel, MassScalingChargeStateScoringModel,
    ChargeStateDistributionScoringModelBase)

from .chromatogram_solution import (
    ChromatogramSolution, ChromatogramScorer, ModelAveragingScorer,
    CompositionDispatchScorer, DummyScorer, ChromatogramScoreSet,
    ScorerBase)

from .mass_shift_scoring import (
    MassScalingMassShiftScoringModel)

from .utils import logit, logitsum


__all__ = [
    "ScoringFeatureBase", "DummyFeature", "epsilon",

    "ChromatogramShapeFitter", "ChromatogramShapeModel",
    "AdaptiveMultimodalChromatogramShapeFitter",

    "ChromatogramSpacingFitter", "ChromatogramSpacingModel",
    "PartitionAwareRelativeScaleChromatogramSpacingFitter",
    "total_intensity",

    "IsotopicPatternConsistencyFitter", "IsotopicPatternConsistencyModel",
    "UniformChargeStateScoringModel", "MassScalingChargeStateScoringModel",
    "ChargeStateDistributionScoringModelBase",

    "ChromatogramSolution", "ChromatogramScorer", "ModelAveragingScorer",
    "CompositionDispatchScorer", "DummyScorer", "ChromatogramScoreSet",
    "ScorerBase",

    "MassScalingMassShiftScoringModel",

    'logit', 'logitsum',
]
